#ifndef BP2WBASE_H
#define BP2WBASE_H

#ifdef WIN32

class BppVirtualMachine;
class ValPtr;

#include <bppstrng.h>
#include <windows.h>
#ifdef WINCE
  #include <winuser.h>
#endif

/** @addtogroup Bp2ClassLib
@{
*/

/** @addtogroup Bp2WinBase Windows Base Classes
@{
*/

#ifdef DOXY
/** Base class of installable message handlers.

*/
class WinMsgHandler
{ 
/** @publicsection */

   WinMsgHandler(WinBase owner = null);
   ~Bp2WinMsgHandler();
   null setOwner(WinBase owner);
   WinBase getOwner();
   int handleMessage(WinBase sender, int hwnd, int msgId, int wParam, int lParam);
   int equals();
/** @protectedsection */
   WinBase _owner;
};


// #endif // DOXY


/** Base class of all window objects.
This class is intended to be used for base of all windows, i.e.
all classes having its own window handle. It provides a set of functions common to
all windows as well as basic window and application management services.
*/
class WinBase
{
public:
   /// ctor
   WinBase();
   /// dtor
   ~WinBase();
   /// create wrapper for existing window
   int createFromHandle(int hwnd);
   /// create window from window class
   int create(string wndClass, int hwndParent = 0, int x = -1, int y = -1, int width = 100, int height = 100, 
       int style = 0, int extStyle = 0, int hMenu_ID = 0, string text = "");
   /// get associated window handle
   int getHandle();
   /// get handle of parent window
   int getParentHandle();
   /// get parent window
   WinBase getParent();
   /// add child window to this window
   int addChild(WinBase child);
   /// get list of all child window handles
   vector getChildrenHandles();

   /// set window position relative to parent
   int setPosition(int x, int y, int repaint = 1);
   /// set window position relative to parent
   int setPosition(POINT pos, int repaint = 1);

   /// set absolute size of window in pixels
   int setSize(int width, int height, int repaint = 1);
   /// set absolute size of window in pixels
   int setSize(SIZE size, int repaint = 1);

   /// get position and size of window
   RECT getWindowRect(); 
   /// get position and size of client area
   RECT getClientRect(); 
   /// get window position relative to parent
   POINT getPosition();
   /// get position of client area
   POINT getClientPos();
   /// get absolute size in pixels
   SIZE getSize();
   /// get absolute size of client area in pixels
   SIZE getClientSize();
   /// get x position (horizontal position)
   int getPosX();
   /// get y position (vertical position)
   int getPosY();
   /// get x position of client area
   int getClientPosX();
   /// get y position of client area
   int getClientPosY();
   /// get window width
   int  getWidth();
   /// get window height
   int getHeight();
   /// get width of client area
   int getClientWidth();
   /// get height of client area
   int getClientHeight();
 
   /// get window style
   int getStyle() const;
   /// check if window has specified style(s)
   int hasStyle(int style);
   /// set window style
   int setStyle(int style);
   /// add new style(s) to window
   int addStyle(int style);
   /// remove specified style(s) from window
   int removeStyle(int style);

   /// get extended window style
   int getExtStyle();
   /// check if window has specified extended style(s)
   int hasExtStyle(int style);
   /// set extended window style
   int setExtStyle(int style);
   /// add new extended style(s) to window
   int addExtStyle(int style);
   /// remove specified extended style(s) from window
   int removeExtStyle(int style);

   /// check if window currently minimized
   int isMinimized();
   /// check if window currently maximized
   int isMaximized();
   /// check if window currently visible
   int isVisible();
   /// minimize window
   int minimize();
   /// maximize window
   int maximize();
   /// show window in specified mode
   int show(int cmdShow = SW_SHOWNORMAL);
   /// restore normal default of window
   int restore();
   /// hide window
   int hide();

   /// message preprocessing function
   int preprocessMessage(MSG msg);
   /// message handling function
   int handleMessage(int hwnd, int msgId, int wParam, int lParam);

   /// add installable message handler to window
   int addMsgHandler(WinMsgHandler handler);
   /// remove specified message handler from window
   int removeMsgHandler(WinMsgHandler handler);
   /// remove all installable message handlers from window
   null removeAllMsgHandlers();

   /// get array of accelerator definitions
   vector getAcceleratorArray();
   /// add accelerator definition to window
   null addAccelerator(int key, int cmd, int flags = 0);
   /// apply all accelerator definitions to window
   null applyAccelerators();

   /// register a window
   static int registerWindow(WinBase pwin);
   /// unregister a window
   static int unregisterWindow(int hwnd);
   /// find window by its handle
   static WinBase findWindow(int hwnd);
   /// force termination of message loop
   static null forceQuit();
   /// set automatic quit mode
   static null setAutoQuit(int autoQuit);
   /// set applications return code
   static null setReturnCode(int returnCode);

   /// set text of window (e.g. used for caption)
   int setWindowText(var title);
   /// get text of window
   string getWindowText();
   /// set color of window background
   int setBackgroundColor(int backColor);
   /// set brush for drawing window background
   int setBackgroundBrush(int backBrush);
   /// set foreground color (e.g. usable for drawing text)
   int setForegroundColor(int foreColor);
   /// implementation of application main loop
   static null mainLoop();
   /// set module instance handle of bp2class (internal used)
   static null setHInstance(int hInstance);
   /// get module instance handle of bp2class (needed e.g. for resource access)
   static int getHInstance();
   /// helper for ID creation (for internal use)
   static int createID();
   /// helper for initialization of common controls
   static int initCommonControls(int dwICC);
   //enum KSWINLIB_EXPORT ScrollBarMode
   //{
   //   SbNone = 0,
   //   SbVertical = 1,
   //   SbHorizontal= 2,
   //   SbBoth = SbVertical | SbHorizontal,
   //   SbAuto = 4
   //};
   int enableVirtualSize(int enable = 1);
   null setLifeScroll(int lifeScroll = 1);
   int setVirtualSize(int width, int height);
   int setVirtualSize(SIZE size);
   SIZE getVirtualSize();
   int fitVirtualSizeToContents(int borderWidth = 0);
   POINT getVirtualOrigin();
   int scrollHorizontal(int offset);
   int scrollVertical(int offset);
   bool setScrollBarMode(int scrollBarMode);
   null setScrollSteps(int sx=10, int sy=10);
   int getScrollBarMode() const;
   virtual null applyScrollBars();

   /** Starts Paint Operation
   @return Handle of device context on success (as intptr), otherwise 0
   @note Method may be used inside handling of WM_PAINT messages in order to get
   paint DC for window. When painting operation done, client code must call endPaint().
   @see endPaint
   */
   HDC beginPaint();

   /** Ends Paint Operation
   @return Nonzero value on success.
   @note Method is used handling of WM_PAINT messages in order to end operation.
   @see beginPaint
   */
   int endPaint();

protected:
   intptr _hwnd; ///< window handle
   vector _handlerList; ///< list of installed message handlers
   intptr _savedWndProc; ///< original window procedure of associated window

   RECT _virtualRect;
   int _useVirtualSize;
   int _scrollBarMode;
   int _sx;
   int _sy;
   int _lifeScroll;

   int _foreColor; ///< current foreground color (COLORREF)
   int _backColor; ///< current background color (COLORREF)
   int _backBrush; ///< current background brush handle
   static int _maxID; ///< maximum ID used (helper for createID)
   static WinBase _pWinCreate; ///< helper variable for temporary store window reference during creation


   /// array of accelerator definitions
   vector _accelArray;
   /// handle of accelerator table currently in use
   intptr _haccel;
   /// handle of internal PAINTSTRUCT currently in use
   intptr _paintStruct; 
   /// module instance handle of KsWinLib
   static HINSTANCE _hInstance;
   /// return code when application exits
   static int _returnCode;
   /// auto quit flag
   static int _autoQuit;
};

#endif // DOXY
/// @}
#ifndef DOXYUSER
/** @addtogroup Bp2Impl Implementation
@{
*/

/** @name WinBaseImpl Windows base class implementation */
//@{

class BppObject;

class Bp2WinMsgHandler
{
public:
   static void CreateClass(BppVirtualMachine& vm);
   static int _hdlOwnerIdx;
   static const BppString _className;
private:
   static bool ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setOwner(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getOwner(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool handleMessage(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool equals(BppVirtualMachine* vm, int argc, ValPtr& sp);
   Bp2WinMsgHandler();
   ~Bp2WinMsgHandler();
};

class Bp2WinBase
{
public:
   static void CreateClass(BppVirtualMachine& vm);
   static const BppString _className;
   static int _hwndIdx; ///< window handle
   static int _handlerListIdx; ///< list of installed message handlers
   static int _savedWndProcIdx; ///< original window procedure of associated window

   static int _virtualRectIdx;
   static int _useVirtualSizeIdx;
   static int _scrollBarModeIdx;
   static int _sxIdx;
   static int _syIdx;
   static int _lifeScrollIdx;

   static int _foreColorIdx; ///< current foreground color (COLORREF)
   static int _backColorIdx; ///< current background color (COLORREF)
   static int _backBrushIdx; ///< current background brush handle


   /// array of accelerator definitions
   static int _accelArrayIdx;
   /// handle of accelerator table currently in use
   static int _haccelIdx;
   static int _paintStructIdx; ///< PAINTSTRUCT handle

   static const BppString _maxIDNm; ///< maximum ID used (helper for createID)
  /// return code when application exits
   static const BppString _returnCodeNm;
   /// auto quit flag
   static const BppString _autoQuitNm;

   static void setHInstance(HINSTANCE hInstance);
   static HINSTANCE getHInstance();
   /// global message distribution callback
   static LRESULT CALLBACK globalMsgCallback(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

private:
   static bool ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool dtor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool createFromHandle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool create(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getHandle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getParentHandle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getParent(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool addChild(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getChildrenHandles(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setPosition(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getWindowRect(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientRect(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getPosition(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientPos(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getPosX(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getPosY(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientPosX(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientPosY(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getWidth(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getHeight(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientWidth(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientHeight(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool hasStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool addStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool removeStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getExtStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool hasExtStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setExtStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool addExtStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool removeExtStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool isMinimized(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool isMaximized(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool isVisible(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool minimize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool maximize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool show(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool restore(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool hide(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool preprocessMessage(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool handleMessage(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool addMsgHandler(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool removeMsgHandler(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool removeAllMsgHandlers(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getAcceleratorArray(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool addAccelerator(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool applyAccelerators(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool registerWindow(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool unregisterWindow(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool findWindow(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool forceQuit(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setAutoQuit(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setReturnCode(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setWindowText(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getWindowText(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setBackgroundColor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setBackgroundBrush(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setForegroundColor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool mainLoop(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setHInstance(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getHInstance(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool createID(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool enableVirtualSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setLifeScroll(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setVirtualSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getVirtualSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool fitVirtualSizeToContents(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getVirtualOrigin(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool scrollHorizontal(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool scrollVertical(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setScrollBarMode(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setScrollSteps(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getScrollBarMode(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool applyScrollBars(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool beginPaint(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool endPaint(BppVirtualMachine* vm, int argc, ValPtr& sp);

   static bool initCommonControls(BppVirtualMachine* vm, int argc, ValPtr& sp);

   /// module instance handle of this library
   static HINSTANCE _hInstance;

   static BppObject* _pWinCreate;
};
//@}
/// @}
#endif //!DOXYUSER
/// @}

#endif // WIN32
#endif // !BP2WBASE_H
