#ifndef BPPBUF_H
#define BPPBUF_H
#include <bpprobj.h>

class BppDictionaryIO;

/** Byte buffer class.
This class provides a buffer, i.e. a container of any data. It is used for representation
of buffer values and as a inernal helper.
*/
class BPPLIB_DLLEXP BppBuffer : public BppRefObject
{
	friend class BppDictionaryIO;
public:
	/// default ctor
	BppBuffer();
	/// ctor with specified size
	BppBuffer(size_t capacity);
	/// copy ctor
	BppBuffer(const BppBuffer& cpy);
	/// dtor
	virtual ~BppBuffer();
		/// create clone of current object
	BppBuffer* clone() const;

	/// assignment
	BppBuffer& operator=(const BppBuffer& assignment);
	/** Access byte via index.
	@param index Index of element to access to. The index must be lower than current capacity.
	@return writable reference to specified element
	@note Operator directly accesses the internal data buffer without any range checking.
	*/
	inline BppByte& operator[](size_t index)
	{
		return _data[index];
	}

	/** Read byte via index.
	@param index Index of element to access to. The index must be lower than current capacity.
	@return value of specified element
	@note Operator directly accesses the size_ternal data buffer without any range checking.
	*/
	inline BppByte operator[](size_t index) const
	{
		return _data[index];
	}   

	/** get current size
	@return number of elements (bytes) currently managed by the buffer
	*/
	inline size_t size() const BPPNOEXCEPT
	{
		return _size;
	}

	/** get current capacity
	@return maximum number of bytes buffer can hold without resizing
	*/
	inline size_t capacity() const BPPNOEXCEPT
	{
		return _capacity;
	}

	size_t setCapacity(size_t capacity);
	/// shrink capacity to current size
	size_t shrink();
	/// concat two buffers
	BppBuffer operator+(const BppBuffer& app);
	/// append content of another buffer to this buffer
	BppBuffer& operator+=(const BppBuffer& app);
	/// get and remove lase byte from buffer
	BppByte pop_back();
	/// get and remove last word (16 bits) drom buffer
	BppWord pop_backW();
	/// get and remove last dword (32 bits) from buffer
	BppDWord pop_backDW();
	/// get and remove last native word from buffer
	BppNWord pop_backNW();
	
	/// get word (16 bits) at specified position
	BppWord getW(size_t pos) const;
	/// get dword (32 bits) at specified position
	BppDWord getDW(size_t pos) const;
	/// get native word (size depends on platform) at specified position
	BppNWord getNW(size_t pos) const;
	/// set word (16 bits) at specified position
	void setW(size_t pos, BppWord val);
	/// set dword (32 bits) at specified position
	void setDW(size_t pos, BppDWord val);
	/// set native word (size depends on platform) at specified position
	void setNW(size_t pos, BppNWord val);

	/// push byte to end of buffer
	size_t push_back(BppByte b);
	/// push word (16 bits) to end of buffer
	size_t push_backW(BppWord w);
	/// pusch dword (32 bits) toend of buffer
	size_t push_backDW(BppDWord dw);
	/// push native word to end of buffer
	size_t push_backNW(BppNWord nw);
	/// clear buffer
	void clear();
	/** Gets buffer data as a C-array.
	@return pointer to internal data area
	@note This function is mainly for internal use. Client code must not delete nor resize returned data space.
	*/
	BppByte* data() const
	{
		return _data;
	}   
	/// set buffer size
	bool setSize(size_t size);
	/// compare function
	int compare(const BppBuffer& comparison) const BPPNOEXCEPT;
	/// equality operator
	bool operator==(const BppBuffer& cmp) const BPPNOEXCEPT;
	/// unequality operator
	bool operator!=(const BppBuffer& cmp) const BPPNOEXCEPT;
private:
	/// current size (number of bytes  contained in buffer)
	size_t _size;
	/// currently available space (bytes) in buffer
	size_t _capacity;
	/// internal data pointer
	BppByte* _data;
};

#ifndef DOXY
BppDWord BytePtr2DW(BppByte* p);
BppWord BytePtr2W(BppByte* p);
BppNWord BytePtr2NW(BppByte* p);
#endif

/** Character buffer class.
This class is very similarly to BppBuffer class. However, type of values
inside the buffer is a character.
*/
class BPPLIB_DLLEXP BppCharBuffer : public BppRefObject
{
public:
	/// default ctor
	BppCharBuffer();
	/// ctor with initially capacity setting
	BppCharBuffer(size_t capacity);
	/// ctor with initialization fron C string
	BppCharBuffer(const BppChar* cstr);
	/// copy ctor
	BppCharBuffer(const BppCharBuffer& cpy);
	/// assignment
	BppCharBuffer& operator=(const BppCharBuffer& assignment);
	/// assignment from C string
	BppCharBuffer& operator=(const BppChar* cstr);
	/// dtor
	virtual ~BppCharBuffer();
	/// create clone of current object
	BppCharBuffer* clone() const;
	/// read/write access to single character by index
	BppChar& operator[](size_t index);
	/// get single character by index
	BppChar operator[](size_t index) const;
	/// get current size
	size_t size() const BPPNOEXCEPT;
	/// get current capacity
	size_t capacity() const BPPNOEXCEPT;
	// set capacity
	size_t setCapacity(size_t capacity);
	/// shrink capacity to current size
	size_t shrink();
	/// concatenate contents of two buffers
	BppCharBuffer operator+(const BppCharBuffer& app);
	/// append contents of specified buffer to end of current one
	BppCharBuffer& operator+=(const BppCharBuffer& app);
	/// get and remove last character from buffer
	BppChar pop_back();
	/// append character to end of buffer
	void push_back(BppChar b);
	/// clear buffer
	void clear();
	/// get buffer data as a C-array
	BppChar* data() const BPPNOEXCEPT;
	/// make a valid C-string from buffer and set _size to length of string
	void adjustSize();
	/// compare function
	int compare(const BppCharBuffer& comparison) const BPPNOEXCEPT;
	/// equality operator
	bool operator==(const BppCharBuffer& cmp) const BPPNOEXCEPT;
	/// unequality operator
	bool operator!=(const BppCharBuffer& cmp) const BPPNOEXCEPT;
private:
	/// size of buffer
	size_t _size;
	/// capacity of buffer
	size_t _capacity;
	/// internal data pointer
	BppChar* _data;
};

#endif // BPPBUF_H
