// Documentation only - don't include it

/** @addtogroup Bp2Functions B++ predefined functions
@{
*/

/** @addtogroup ContainerFuncs Container related functions
@{
*/

/** Clones a value of any type.
Function creates a clone of <i>val</i>. It may be used to create full (deep) copy of
any container type, string or object instance. For all other types (including
references) it simply returns argument itself.
@param val Value to clone.
@return cloned value
*/
var clone(var val);

/** Create a new vector.
Function creates a new vector and optionally initializes its elements.
@param capacity initial capacity of vector
@param init Flag for vector initialization. If <tt>true</tt> (i.e. not 0),
vector is initialized, i.e. vector size is set to be same as <i>capacity</i>
and all elements are set to <tt>null</tt>.
@return vector object
@note There is currently no way to change capacity of vector in B++ code
explicitely after creation. Setting capacity and initialization on creation
may increase speed because there is no need to resize vector at runtime.
@par Usage example:
@code
vec = newvector(10);
for (i=1;i<=10;++i) vec[i-1] = i*i;
@endcode
@see vinsert, verase
*/
vector newvector(int capacity=0, int init=1);

/** Create buffer as a copy of another buffer or string.
Function makes a complete copy of <i>assignment</i> value.
@param assignment Value to create new buffer from. Value must be of type
<tt>string</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@return new buffer
@note In UNICODE builds wide characters will be converted into single bytes
discarding high byte of each character.
*/
buffer newbuffer(var assignment);
/** Create buffer filled with a specified byte.
@param size Number of bytes in buffer to be created.
@param fillbyte Byte value to initialize buffer with.
@return new buffer
*/
buffer newbuffer(int size, int fillbyte=0);


/** Create character buffer as a copy of another buffer or string.
Function makes a complete copy of <i>assignment</i> value.
@param assignment Value to create new character buffer from. Value must be of type
<tt>string</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@return new character buffer
@note In UNICODE builds if <i>assignment</i> is of type <tt>buffer</tt>
each byte of <i>assignment</i> is converted into wide character
setting zero for high byte.
*/
charbuffer newcbuffer(var assignment);


/** Create character buffer filled with a specified character.
@param size Number of bytes in buffer to be created.
@param fillchar Character value to initialize buffer with.
@return new character buffer
*/
charbuffer newcbuffer(int size, int fillchar=0);

/** Create new empty dictionary.
@param sorted Flag for sorted dictionary. If non zero entries in dictionary will
be ordered by keys, otherwise new elements are appended to end.
@return new dictionary
@note <ul><li>Value of <i>sorted</i> flag affects results of dictgetkeys and dictgetvalues
functions. If set these functions return entries key ordered, otherwise in order of
insertion.</li>
<li>In DOS and WindowsCE builds element access will be faster in sorted than non-sorted
dictionaries.</li>
<li>Dictionaries created by literal constructor ({...}) are not sorted.</li>
</ul>
*/
dictionary newdict(int sorted = 0);

/** Check for dictionary entry.
Function checks if an entry for <i>key</i> is contained in <i>dict</i>.
@param dict dictionary to search
@param key key to search for
@return 1 if found, otherwise 0
*/
int dictcontains(dictionary dict, string key);

/** Remove entry from dictionary.
Function removes entry of key <i>key</i> from <i>dict</i> if contained.
@param dict dictionary to remove entry from
@param key key of entry to remove
@return 1 on success, otherwise 0
*/
int dicterase(dictionary dict, string key);

/** Remove all entries from specified dictionary.
@param dict dictionary to remove entries from
@return null
*/
null dictclear(dictionary dict);

/** Get vector of all keys contained in specified dictionary.
@param dict dictionary to get keys from
@return vector containing all keys of <i>dict</i>
@note <p>Returned vector contains copies of keys in dictionary. Therefore client code can modify
strings in returned vector without sideeffects for <i>dict</i>.</p>
<p>Returned vector may be used instead of an iterator to walk through the dictionary.</p>
<p>Order of keys in returned vector depends on sorted state of dictionary. In a sorted
dictionary keys are in alphabetical order, otherwise in order of insertion.
@par Usage example;
@code
   dictionary myDict = newdict();  // create unsorted dictionary
   // add any entries here
   // ...
   vector keyvec = dictgetkeys(myDict);
   uint n = size(keyvec);
   for (uint i=0; i<n;++i)
      print(keyvec[i],":",myDict[keyvec[i]],"\n");
@endcode
@see newdict, dictgetvalues
*/
vector dictgetkeys(dictionary dict);

/** Get vector of all values contained in specified dictionary.
@param dict dictionary to get values from
@return vector containing all values of <i>dict</i>
@note
<p>Returned vector may be used instead of an iterator to walk through the dictionary.</p>
<p>Order of values in returned vector conforms to order of keys.</p>
@par Usage example;
@code
   dictionary myDict = newdict();  // create unsorted dictionary
   // add any entries here
   // ...
   vector keyvec = dictgetkeys(myDict);
   vector valvec = dictgetvalues(myDict);
   uint n = size(keyvec); // same as size(valvec)
   for (uint i=0; i<n;++i)
      print(keyvec[i],":",valvec[i],"\n");
@endcode
@see newdict, dictgetkeys
*/
vector dictgetvalues(dictionary dict);

/** Get number of elements in a container.
@param container Any container, i.e. value of type <tt>buffer</tt>,
<tt>charbuffer</tt>, <tt>vector</tt> or <tt>dictionary</tt>, or <tt>string</tt>.
@return number of elements in <i>container</i>
*/
uint size(var container);

/** Shrinks capacity of container to current size.
@param container Value of type <tt>buffer</tt>,
<tt>charbuffer</tt> or <tt>vector</tt>.
@return Capacity (number of elements) of <i>container</i>
*/
uint shrink(var container);

/** Get number of elements in a vector.
@param vec vector to get size from
@return number of elements in <i>vec</i>
@see size
*/
uint vecsize(vector vec);

/** Sorts content of a container using QuickSort algorithm.
Function behaves like qsort function of C library.
@param container Container to sort. <i>container</i> may be of
type <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>vector</tt>.
@param callback Compare callback. The callback may be either
a function or a functor instance (an object defining operator()).
The callback is invoked from qsort algorithm with two arguments (elements
of <tt>container</tt> to compare. It should return a negative number
if left argument is lower than right argument, a positive number if left
argument is greater than right argument or zero in case of equality.<br>
If <i>callback</i> is not specified, comparison is done using default rules.
@return null
@par Usage example;
@code
	class comparer
	{
	   comparer() {}
	   operator()(a,b) { return b-a; } // makes descending order
	};

	void main()
	{
		var vec = [3,1,5,7,2,0,9];
		var sz = size(vec);
		// sort ascending using anonymous function for callback
		qsort(vec,function(a,b) { return a-b; });
		// prints 0123579
		for (var i=0;i<sz;++i) print(vec[i]);
		print("\n");
		// sort descending using comparer object
		qsort(vec,new comparer());
		// prints 9753210
		for (i=0;i<sz;++i) print(vec[i]);
	}
@endcode
*/
null qsort(var container, var callback = null);

/** Enumerates a container and calls callback for each of its items.
@param container Container to enumarate. <i>container</i> may be of
type <tt>buffer</tt>, <tt>charbuffer</tt>,<tt>vector</tt>, <tt>dictionary</tt>
<tt>string</tt> or any class type defining its own <tt>foreach</tt> method of signature<br>
<tt>uint foreach(var callback, ...)</tt>.
@param callback Callback for enumeration. The callback may be either
a function or a functor instance (an object defining operator()).
The callback is invoked from <tt>foreach</tt> with at least two arguments (key and value
of current container element). It should return an integer. If result of <i>callback</i>
is zero (means <tt>false</tt>), enumeration is cancelled.
@param ... If <tt>foreach</tt> called with more than two arguments, additional argumants are
passed to <tt>callback</tt>.
@return Number of items enumerated.
@note <ul>
<li>If <i>container</i> is of <tt>dictionary</tt> type, first argument of <i>callback</i> is
the key of item. For all other types, first argument is the index in <i>container</i>.</li>
<li>Second argument of <i>callback</i> is the value of container item. If item value is of
a reference type, <i>callback</i> can modify it if needed.</li>
<li><i>callback</i> shouldn't modify size of <i>container</i> of enumerated items. If it does
so, it must return <tt>false</tt> to cancel enumeration.</li>
<li>If <i>container</i> is an object, function calls <tt>foreach</tt> method of that
object, passing <i>callback</i> and all additional arguments.
</ul>
@par Usage example;
@code
void main()
{
   // fill sorted dictionary from an unsorted one
   auto dict = { t:'t', e:'e',m:'m',p:'p',o:'o'}; // unsorted
   dictionary dict2 = newdict(true); // sorted
   // use dict2 for additional element
   foreach(dict,
           function(i,v,d) { d[i] = v; return true; },
           dict2);
   // callback for print container items
   auto iter = function(i, v) { print("I: ",i," V: ",v,"\n"); return true;};
   foreach(dict,iter); // list dict
   foreach(dict2,iter); // list dict2
   // enumerate a string
   foreach("Hello World",iter);
}
@endcode
*/
uint foreach(var container, var callback, ...);

/** Finds position of value in vector.
Function searches in <i>cont</i> for value equal to <i>searchVal</i> and returns
position of its first occurrence.
@param cont container to search in
@param searchVal value to search for
@param pos position to start search from.
@return position of first occurrence if found, otherwise -1
@note Function does low level compare, i.e. it doesn't call equals- or compare-methods for objects.
*/
int find(vector cont, var searchVal, int pos = -1);

/** Finds position of value in vector.
Function searches backwards in <i>cont</i> for value equal to <i>searchVal</i> and returns
position of its last occurrence.
@param cont container to search in
@param searchVal value to search for
@param pos position to start backwards search from.
@return position of last occurrence if found, otherwise -1
@note Function does low level compare, i.e. it doesn't call equals- or compare-methods for objects.
*/
int rfind(vector cont, var searchVal, int pos = -1);



/** Create a vector from n-tupel of any values.
Function creates a vector from argument list.
@param ... list of any values
@return vector created from values
@note Other than in vector constructor ([...]) arguments of this function
are not restricted to literals.
@par Usage example;
@code
   string str = "Hello World!\n";
   vector vec = T(str,"This is B++\n",strlen(str);
   uint n = size(vec);
   for (uint i=0;i<n;++i) print(vec[i]);
@endcode
*/
vector T(...);

/** Create a vector from n-tupel of any values.
Function is a synonym to T(...) function.
@param ... list of any values
@return vector created from values
@see T(...)
*/
vector Vec(...);

/** Create a character buffer from n-tupel of values.
Function builds a character buffer from argument list.
@param ... List of values. Elements of argument list
may be of types <tt>int</tt> (interpreted as character), <tt>buffer</tt>,
<tt>charbuffer</tt> or <tt>string</tt>. Arguments of other types are ignored.
@return character buffer created from values
@par Usage example;
@code
   string str = "Hello World!";
   charbuffer myCBuf = CBuf(str,0x20,"Let's ",'b','u','i','l',"d a buffer!",'\n');
   print(string(myCBuf));
@endcode
*/
charbuffer CBuf(...);

/** Inserts value one or more times into vector at specified position.
@param dst vector to insert value into.
@param pos Position in <i>vec</i> where insertion is done. If <i>pos</i> is negative
or greater than maximum valid index in <i>vec</i>, <i>insertion</i> is appended.
@param insertion Value (of any type) to insert. If <i>vec</i> itself is used for
<i>insertion</i> argument, value is cloned before insertion.
@param cnt Number of elements to insert.
@return modified vector
*/
vector vinsert(vector dst, int pos, var insertion, uint cnt = 1);

/** Erases one or more element(s) from vector.
@param vec vector to erase element(s) from.
@param pos Position index of first element to erase.
@param cnt Maximum number of elements to erase. 
@return modified vector
*/
vector verase(vector vec, uint pos = 0, int cnt = -1);

/** Exchanges specified elements of vector.
@param vec vector containing elements to exchange.
@param pos1 Position index of first element.
@param pos2 Position index of second element.
@return modified vector.
@note Function throws exception of string type in case of invalid index.
@par Example;
@code
   vector arr = [1,2,3,4,5,6];
   vsweap(arr,4,1);
   // arr is now [1,5,3,4,2,6]
@endcode
*/
vector vswap(vector vec, uint pos1, uint pos2);

/** Moves an element inside a vector.
@param vec vector containing element to move.
@param srcpos Position index of element to move.
@param dstpos Destination position index for move operation.
@return modified vector.
@note Function throws exception of string type in case of invalid index.
@par Example
@code
   vector arr = [1,2,3,4,5,6];
   vmove(arr,4,1);
   // arr is now [1,5,2,3,4,6]
@endcode
*/
vector vmove(vector vec, uint srcpos, int udstpos);

/** Extracts a subsequenze of elements from vector.
@param vec vector to build subsequence from.
@param from Index of first element to extract in <i>vec</i>.
@param cnt Number of elements to extract. If <i>cnt</i> is negative or greater than 
number of elements behind <i>from</i>, all elements to end of <i>vec</i> are icluded into result.
@return New vector containing selected Elements of <i>vec</i>.
*/
vector vslice(vector vec, uint from = 0, int cnt = -1);

/** Removes all elements from specified vector.
@param vec vector to clear,
*/
void vclear(vector vec);

/// @}

/** @addtogroup StringFuncs String functions
@{
*/
/** Create string as a copy of another string or buffer.
Function makes a complete copy of <i>assignment</i> value.
@param assignment Value to create new string from. Value must be of type
<tt>string</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>. In case of
<tt>buffer</tt> each byte contained in <i>assignment</i> is assumed to
be a single byte character.
@return new string
*/
string newstring(var assignment);

/** Create string from format specification.
Function builds new string from format specification and any number
of additional arguments.
@param formatstr Format specification. Value must be of <tt>string</tt> type.
The format string has similary syntax as in C - see BppString description.
@param ... More arguments to replace variables in <i>formatstring</i>.
@return new string
@note Elements of variable argument list normally should be of simple types, i.e.
integer, double, float or string. For all other types
return value of BppValue::addr() method is used for variable value.
*/
string newstring(string formatstr,...);

/** Creates string filled with a specified character.
@param size Number of characters in string to be created (without trailing '\\0'
character.
@param fillchar Character to initialize string with.
@return new string
*/
string newstring(uint size, int fillchar=' ');

/** Get length of a string.
@param str <tt>string</tt> value to get length of
@return number of characters in <i>str</i> (without trailing zero)
*/
uint strlen(string str);

/** Compare function (case sensitive).
@param str1 left <tt>string</tt> for compare
@param str2 right <tt>string</tt> for compare
@return <ul>
<li> &lt; 0 if <i>str1</i> is lower than <i>str2</i></li>
<li>  0 if <i>str1</i> equals to <i>str2</i></li>
<li> &gt; 0 if <i>str1</i> is greater than <i>str2</i></li>
</ul>
*/
int strcmp(string str1, string str2);

/** Compare function (case insensitive).
@param str1 left <tt>string</tt> for compare
@param str2 right <tt>string</tt> for compare
@return <ul>
<li> &lt; 0 if <i>str1</i> is lower than <i>str2</i></li>
<li>  0 if <i>str1</i> equals to <i>str2</i></li>
<li> &gt; 0 if <i>str1</i> is greater than <i>str2</i></li>
</ul>
*/
int stricmp(string str1, string str2);

/** String tokenize function.
Function splits <i>str</i> into tokens using delimiters defined in <i>delimiter</i>.
@param str <tt>string</tt> to tokenize
@param delimiter Set of characters (type <tt>string</tt>) to be used for delimiters
@param storeEmptyTokens If <tt>true</tt>, empty tokens are stored in result too.
@param storeDelimitChar If <tt>true</tt>, delimiter character between tokens are stored in result too.
@return vector containing tokens
@note <ul>
  <li>If <i>str</i> is empty, result vector is empty too.</li>
  <li>Delimiters (if <i>storeDelimitChar</i> set) are atored in result as character codes (i.e. of type <tt>int</tt>).</li>
  <li>If <i>storeEmptyTokens</i> is set and last character of <i>str</i> is a delimiter, empty
  string is appended to result.</i>
 </ul>
@par Usage example;
@code
void listvec(vector v)
{
   int n = size(v);
   print("size: ",n,"\n");
   for (int i=0;i<n;++i) print(""+v[i],"\n");
   print("-----\n");
}

main() {
   string str = "This is a  sample string.";
   listvec(strsplit(str," ."));
   listvec(strsplit(str," .",false,true));
   listvec(strsplit(str," .",true));
   listvec(strsplit(str," .",false,true));
   listvec(strsplit(str," .",true,true));
}
@endcode
*/
vector strsplit(string str, string delimiter, bool storeEmptyTokens = false, bool storeDelimitChar= false);

/** Find position of substring or character.
Function searches <i>str</i> for substring or character <i>search</i> and returns
position of its first occurrence.
@param str string to search
@param search string or character to search for
@param start character position to start search from.
@return position of first occurrence if found, otherwise -1
*/
int strfind(string str, var search, uint start = 0);

/** Erase part of string.
Function deletes one or more characters from string.
@param str string to erase character(s) from
@param pos index of first character to delete
@param cnt Number of characters to delete. If a negative value is specified
for <i>cnt</i> all characters from <i>pos</i> to end of string will be deleted.
@return modified string
@note Function modifies argument <i>str</i>. The return value is the
modified string itself, not a copy of it.
*/
string strerase(string str, uint pos, int cnt = -1);

/** Insert string into another one.
Function inserts value of <i>insstr</i> into <i>str</i> at position <i>pos</i>.
@param str string to modify
@param pos Index position where <i>insstr</i> should be inserted. Use length
of <i>str</i> for <i>pos</i> to append <i>insstr</i> value.
@param insstr string to insert
@return modified string
@note Function modifies argument <i>str</i>. The return value is the
modified string itself, not a copy of it.
*/
string strinsert(string str, uint pos, string insstr);

/** Insert character(s) into string.
Function inserts <i>cnt</i> characters <i>insc</i> into <i>str</i> at position <i>pos</i>.
@param str string to modify
@param pos Index position for insertion. Use length
of <i>str</i> for <i>pos</i> to append characters.
@param insc character value to insert
@param cnt number of characters to insert
@return modified string
@note Function modifies argument <i>str</i>. The return value is the
modified string itself, not a copy of it.
*/
string strinsert(string str, uint pos, int insc, uint cnt = 1);

/** Replace part of string.
Function replaces all occurrences of <i>search</i> by value of <i>replacement</i>.
@param str string to modify
@param search character or substring to be replaced
@param replacement character or string to be used for replacement
@return current object
@note Both parameters must be either single characters or strings.
*/
string strreplace(string str, var search, var replacement);

/** Get substring from string.
Function returns substring of <i>str</i>, beginning at position <i>start</i>
and of maximum <i>len</i> characters length.
@param str string to get substrung from
@param start Start position of substring.
@param len Maximum size of substring. If len is omitted of negative,
all characters of current string (beginning at <i>start</i>) are included in result.
@return new string containing specified substring
*/
string substr(string str, uint start, int len = -1);


/** Make string upper case.
Function converts letters contained in str to upper case.
@param str string to convert
@return converted string
@note Function modifies content of <i>str</i> and returns reference to modified string.<br>
It behaves like C-function of same name, i.e. result depends on locale settings.
@see setlocale, strlwr
*/
string strupr(string str);

/** Make string lower case.
Function converts letters contained in str to lower case.
@param str string to convert
@return converted string
@note Function modifies content of <i>str</i> and returns reference to modified string.<br>
It behaves like C-function of same name, i.e. result depends on locale settings.
@see setlocale, strupr
*/
string strlwr(string str);
/// @}

/** @addtogroup MathFuncs Math functions

All functions of this section accept argument(s) of any numeric type. <br>
Return value is (with except of #abs function)
of <tt>double</tt> type.
@note If an argument is of object type, functions (with except of #abs) try to
convert it using double conversion operator of argument.
@{
*/

/** Calculates sin(<i>x</i>).
@param x argument (radiant)
@return sinus of <i>x</i>
*/
double sin(var x);

/** Calculates cos(<i>x</i>).
@param x argument (radiant)
@return cosinus of <i>x</i>
*/
double cos(var x);

/** Calculates tan(<i>x</i>).
@param x argument (radiant)
@return tangens of <i>x</i>
*/
double tan(var x);

/** Calculates arctan(<i>x</i>).
@param x argument
@return arcus tangens of <i>x</i> (radiant)
*/
double atan(var x);

/** Calculates square root.
@param x argument, must not be negative
@return square root of <i>x</i>
*/
double sqrt(var x);

/** Calculates natural logarithm ln(<i>x</i>).
@param x argument, must be positive
@return natural logarithm of <i>x</i>
*/
double log(var x);

/** Calculates expotential function <b>e</b><sup><i>x</i></sup>.
@param x argument
@return <b>e</b> to the power <i>x</i>
*/
double exp(var x);

/** Calculates expotential function <i>x</i><sup><i>y</i></sup>.
@param x base
@param y exponent
@return <i>x</i> to the power of <i>y</i>
*/
double pow(var x, var y);

/** Calculates absolute value of <i>x</i>.
@param x argument
@return absolute value of <i>x</i>
@note Return type is same as type of <i>x</i>.
*/
var abs(var x);

/// @}

/** @addtogroup IOFuncs I/O functions
@{
*/

/** @name Common I/O functions
@note Functions of this section are available in all B++ builds.
*/
//@{

/** Print list of values to default output device.
@param ... Comma separated list of values. Output depends on value type as follows:
<ul>
   <li><tt>string</tt> : Value is written to output without any changes.</li>
   <li>all integer types (<tt>sbyte ... ulong</tt>): Value is written as unformatted integer or unsigned integer (depending on type).</li>
   <li><tt>double</tt> or <tt>float</tt> : Value is written as unformatted 
   double of automatic notation (like %g-format pf printf).</li>
   <li><tt>intptr</tt> : Value is written as address in hexadecimal representation.</li>
   <li>all other types : Type name and address of value are written to output.</li>
</ul>
@return null
@note Output is done using IBppPrinter implementation set for default printer of
virtual machine.
*/
null print(...);


/** Open a file.
Function tries to open file <i>filename</i> in specified open mode. It behaves
exactly as same-named C-function.
@param filename Name of file to open. The name may contain relative or absolute
peth specification.
@param mode Specification of open mode, may contain following elements:
<ul>
   <li>r : Open an existing file for read only.</li>
   <li>w : Create file and open it for write. If specified file already exists it is overwritten.</li>
   <li>a : Open a file for append data. If file doesn't exist it will be created.</li>
   <li>r+ : Open existinng file for read and write.</li>
   <li>w+ : Create file and open it for read and write. If specified file already exists it is overwritten.</li>
   <li>t : Open file in text mode (may be combined with a, r, r+, w and w+).</li>
   <li>b : Open file in binary mode (may be combined with a, r, r+, w and w+).</li>
</ul>
@return File descriptor on success, otherwise <tt>null</tt>.
@note Files opened by this function should be closed using fclose if no longer needed.
@see fclose
*/
FILE fopen(string filename, string mode);

/** Close a file.
Function closes a file previously opened by fopen.
@param fp descriptor of file to be closed
@return 0 on success, otherwise non-zero value
@note Function must not be used with default streams stdin, stdout or stderr.
*/
int fclose(FILE fp);

/** Check for end of file.
Function checks if end of specified file is reached.
@param fp descriptor of file
@return non-zero value if end of file is reached, otherwise 0
*/
int feof(FILE fp);

/** Read single character from opened file.
Function reads a single-byte character from specified file.
@param fp file descriptor
@return character code (-1 if end of file reached)
*/
int getc(FILE fp);

/** Read single character input object installed for virtual machine.
Function reads a single character from input object installed for virtual machine.
@return character code (-1 if nothing read)
@note By default virtual machine uses <tt>stdin</tt> device to get the character.
However, another input object (IBppInput implementation) may be installed by client code.
*/
int getc();

/** Write single character to opened file.
Function writes a single-byte character to specified file.
@param c character to write
@param fp file descriptor
@return character code written (-1 on error)
*/
int putc(int c, FILE fp);

/** Read string from standard input.
Function by default reads string from stdin device (usually the console). It returns
on new line (i.i.erase. Enter key). The trailing line feed is not contained
in return value.
@return string read
@note Function normally works on systems having a console only. However, an input object (IBppInput implemantation)
may be installed for virtual machine to get input from another device than <tt>stdin</tt>.
@see fgets
*/
string gets();

/** Read string from file.
Function reads a string from an opened file, beginning at current position.
If returns on new line or if end of file is reached.
@param fp descriptor of file to read from
@return string read
@note <p>Trailing line feed character is not part of result.</p>
<p>Function should be used only for reading from files opened in text mode.</p>
@see gets, fread
*/
string fgets(FILE fp);

/** Write string to file.
Function writes <i>text</i> into opened file <i>fp</i>, beginning at current position.
@param text string to write
@param fp descriptor of file to write into
@return on success code of last character written, otherwise EOF (-1).
@note Function should be used for writing to files opened in text mode only.
@see fwrite
*/
int fputs(string text, FILE fp);

/** Read value from file.
Function reads any B++ value from opened binary file, beginning at current position.
@param fp Descriptor of file to read from. The file must be in format as created by
fwriteval function.
@return on success value read, otherwise <tt>null</tt>
@note Client code should assign returned value to a variable for further use.
Please refer to #fwriteval documentation for a usage example.
@see fwriteval
*/
var freadval(FILE fp);

/** Write value to file.
Function writes any B++ value into opened binary file, beginning at current position.
@param value value to write
@param fp descriptor of file to write into
@return 1 on success, otherwise 0
@note Using #fwriteval and #freadval functions allows to serialize and deserialize
any B++ data structure in a simple way. Espacially that includes complete object
hierarchies too. Any cross references between objects inside a hierarchy are
saved on serialization and resolved on deserialization.
@par Example:
@code
// Serialization example using fwriteval and freadval.
// first we define a list entry class

class ListEntry
{
   static _maxID = 0;
   _a;
   _b;
   _ID;
   _next;
}

// ctor
ListEntry::ListEntry(a,b)
{
   _ID = ++_maxID;
   _a=a;
   _b=b;
   _next = null;
}

ListEntry::getA() { return _a; }

// print contents of entry
ListEntry::pr()
{
   print("ID: ",_ID," ",getA(), " ",_b,"\n");
}

ListEntry::getNext() { return _next; }

ListEntry::setNext(next) { _next = next; }

ListEntry::append(v)
{
   if (_next)
      _next->append();
   else
   {
      _next = v;
   }
}

// dtor
ListEntry::~ListEntry()
{
   print("destroyed ",_ID,"\n");
}

main()
{
   var lst = new ListEntry(-1,0);
   for(var i=0;i<9;i++)               // append more objects to lst
   {
      lst->append(new ListEntry(i,i+1));
   }
   for(var p = l;p;p=p->getNext())    // print contents
      p->pr();
   var fp=fopen("obj.dat","wb");      // open file for write
   fwriteval(lst,fp);             // write entire list to file
   fclose(fp);
   lst = null;                    // free list
   fp=fopen("obj.dat","rb");      // open file for read
   lst=freadval(fp);              // read contents and assign to lst
   fclose(fp);
   for(p = lst;p;p=p->getNext())  // print contents
      p->pr();
}
@endcode
@see freadval
*/
int fwriteval(var value, FILE fp);

/** Read file.
Function appends contents of file specified by <i>fp</i> to <i>readbuf</i>.
@param readbuf Destination buffer.
@param fp Descriptor of opened file to read from.
@return number of bytes read
@note <ul>
  <li>Contents of <i>fp</i> is read from current file position to end of file.</li>
  <li>Capacity of <i>readbuf</i> is enlarged if neccessary.</li>
</ul>
*/
uint fread(buffer readbuf, FILE fp);

/** Read specified number of bytes from file.
Function appends up to <i>cnt</i> bytes from contents of file specified by <i>fp</i>
to <i>readbuf</i>.
@param readbuf Destination buffer.
@param cnt Maximum number of bytes to read.
@param fp Descriptor of opened file to read from.
@return number of bytes read
@note <ul>
   <li>Capacity of <i>readbuf</i> is enlarged if neccessary.</li>
   <li>Reading starts at current file position</li>
</ul>
*/
uint fread(buffer readbuf,  uint cnt, FILE fp);

/** Read specified number of elements of fixed size from file.
Function appends up to <i>cnt</i> elements of <i>size</i> bytes length from contents
of file specified by <i>fp</i> to <i>readbuf</i>.
@param readbuf Destination buffer.
@param size size of single element to read
@param cnt Maximum number of elements to read.
@param fp Descriptor of opened file to read from.
@return number of elements read
@note <ul>
   <li>Capacity of <i>readbuf</i> is enlarged if neccessary.</li>
   <li>Reading starts at current file position</li>
</ul>
*/
uint fread(buffer readbuf, uint size, uint cnt, FILE fp);

/** Write buffer to file.
Function writes complete contents of <i>buf</i> into file, starting at current
file position.
@param buf Buffer containing contents to write.
@param fp Descriptor of opened file to write to.
@return number of bytes written
*/
uint fwrite(buffer buf, FILE fp);

/** Write specified number of bytes to file.
Function writes up to <i>cnt</i> bytes from <i>buf</i> to file specified by <i>fp</i>.
@param buf Buffer containing contents to write.
@param cnt Maximum number of bytes to write.
@param fp Descriptor of opened file to write to.
@return number of bytes written
@note Function writes into file starting at current file position.
*/
uint fwrite(buffer buf, uint cnt, FILE fp);

/** Write specified number of fixed sized elements to file.
Function writes up to <i>cnt</i> elements of <i>size</i> bytes length from <i>buf</i>
to file specified by <i>fp</i>.
@param buf Buffer containing contents to write.
@param size size of e single element
@param cnt Maximum number of elements to write.
@param fp Descriptor of opened file to write to.
@return number of elements written
@note Function writes into file starting at current file position.
*/
uint fwrite(buffer buf, uint size, uint cnt, FILE fp);

//@}

/** @name Wide character I/O functions
@note Functions of this section are available in Windows builds only.
*/
//@{

/** Read single wide character from opened file.
Function reads a wide character from specified file.
@param fp file descriptor
@return character code (-1 if end of file reached)
*/
int getwc(FILE fp);

/** Write single wide character to opened file.
Function writes a wide character to specified file.
@param c character to write
@param fp file descriptor
@return character code written (-1 on error)
*/
int putwc(int c, FILE fp);

/** Read wide character string from file.
Function reads a wide character string from an opened file, beginning at current position.
If returns on new line or if end of file is reached.
@param fp descriptor of file to read from
@return string read
@note <p>Trailing line feed character is not part of result.</p>
<p>Function should be used only for reading from files opened in text mode.</p>
@see fread
*/
string fgetws(FILE fp);

/** Write wide character string to file.
Function writes <i>text</i> into opened file <i>fp</i>, beginning at current position.
@param text string to write
@param fp descriptor of file to write into
@return on success code of last character written, otherwise EOF (-1).
@note Function should be used only for writing to files opened in text mode.
@see fwrite
*/
int fputws(string text, FILE fp);
//@}

/// @}

/** @addtogroup RTTIFuncs RTTI and type conversion
@{
*/

/** Convert value to boolean value.
Function converts argument of any type to <tt>int</tt> of value 0 (<tt>false</tt>)
or 1 (<tt>true</tt>).
@param value Value to convert.
@return boolean representation for <i>value</i>.<br>
Result is 0 (<tt>false</tt>) for numeric types of value 0, the <tt>null</tt> value and
all reference type having data pointer NULL (i.e. without any data). In all other cases
result value is 1 (<tt>true</tt>), if argument us not an object. For argument of object type
its conversion operators <tt>bool</tt> or <tt>int</tt> are called if defines.
@note 
  Function is mainly intended to prepare a value for comparing with predefined literals <tt>true</tt>
  or <tt>false</tt>.<br>
  For conversion of objects, function searches for <tt>bool</tt> operator first, then
  for <tt>int</tt> operator.
@see size, gettype, int
*/
bool bool(var value);

/** Converts value to signed byte (char).
Function tries to convert argument into value of <tt>sbyte</tt> type (8 bits signed integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>sbyte</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
sbyte sbyte(var value);

/** Converts value to unsigned byte (unsigned char).
Function tries to convert argument into value of <tt>byte</tt> type (8 bits unsigned integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>byte</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
byte byte(var value);


/** Converts value to short integer.
Function tries to convert argument into value of <tt>short</tt> type (16 bits signed integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>short</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
short short(var value);

/** Converts value to unsigned short integer.
Function tries to convert argument into value of <tt>ushort</tt> type (16 bits unsigned integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>ushort</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
ushort ushort(var value);

/** Converts value to integer.
Function tries to convert argument into value of <tt>int</tt> type (32 bits signed integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>int</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
int int(var value);

/** Converts value to unsigned integer.
Function tries to convert argument into value of <tt>uint</tt> type (32 bits unsigned integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>uint</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
uint uint(var value);


/** Converts value to long integer.
Function tries to convert argument into value of <tt>long</tt> type (64 bits signed integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>long</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
long long(var value);

/** Converts value to unsigned long integer.
Function tries to convert argument into value of <tt>ulong</tt> type (64 bits unsigned integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>ulong</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
ulong ulong(var value);

/** Converts value to integer of size same as address pointers of system.
Function tries to convert argument into value of <tt>intptr</tt> type (32 or 64 bits signed integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>intptr</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
intptr intptr(var value);

/** Convert value to <tt>float</tt>.
Function tries to convert argument into value of <tt>float</tt> type.
@param value Value to convert. The value must be of type <tt>int</tt>, <tt>float</tt>,
<tt>double</tt>, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>float</tt>
@note <ul>
<li>A <tt>string</tt> value must be string representation of a <tt>float</tt> number in format<br>
<tt>[<i>whitespace</i>] [<i>sign</i>] [digits] [.digits] [ {d | D | e | E }[sign]digits]</tt><br>
where a whitespace is a space or tab character and sign is either plus (+) or minus (-).</li>
<li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 32 bits (float) or
64 bits (double) value (i.e. size of buffer must be 4 or 8).</li>
<li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>float</tt>
or <tt>double</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
float float(var value);

/** Convert value to <tt>double</tt>.
Function tries to convert argument into value of <tt>double</tt> type.
@param value Value to convert. The value must be of type <tt>int</tt>, <tt>float</tt>,
<tt>double</tt>, <tt>string</tt> or <tt>buffer</tt>.
@return <i>value</i> converted to <tt>double</tt>
@note <ul>
<li>A <tt>string</tt> value must be string representation of a <tt>double</tt> number in format<br>
<tt>[<i>whitespace</i>] [<i>sign</i>] [digits] [.digits] [ {d | D | e | E }[sign]digits]</tt><br>
where a whitespace is a space or tab character and sign is either plus (+) or minus (-).</li>
<li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 32 bits (float) or
64 bits (double) value (i.e. size of buffer must be 4 or 8).</li>
<li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>double</tt>
or <tt>float</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
double double(var value);

/** Convert value to <tt>string</tt>.
Function tries to convert argument into value of <tt>string</tt> type.
@param value Value to convert. The value must be of type <tt>int</tt>, <tt>float</tt>,
<tt>double</tt>, <tt>string</tt>, <tt>charbuffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>string</tt>
@note <ul>
<li>Integer values are converted to decimal string representation.</li>
<li>Values of types <tt>float</tt> and <tt>double</tt> will be converted to string
without any special format (and without expotantial notation).</li>
<li>In case of a <tt>charbuffer</tt> value a new string containing all characters of buffer is returned.</li>
<li>Values of <tt>object</tt> type must have a <tt>string</tt> conversion operator or a
<tt>toString</tt> method for conversion.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
string string(var value);

/** Convert value to <tt>buffer</tt>.
Function tries to convert argument into value of <tt>buffer</tt> type.
@param value Value to convert. The value must be of type <tt>int</tt>, <tt>float</tt>,
<tt>double</tt>, <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>buffer</tt>
@note <ul>
<li>For numeric types (<tt>int</tt>, <tt>float</tt> and <tt>double</tt>) returned buffer
contains a bytewise copy of argument and has same langth as native size of argument type.</li>
<li>In case of a <tt>charbuffer</tt> or <tt>string</tt> all characters of argument, including 
trailing zero, are copied to result buffer. Size of result is number of characters copied
multipled by number of bytes used for a single character.</li>
<li>In case of a <tt>buffer</tt> function simply returns a reference (not a copy!) to argument.</li>
<li>Values of <tt>object</tt> type must have a <tt>buffer</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
buffer buffer(var value);

/** Convert value to <tt>charbuffer</tt>.
Function tries to convert argument into value of <tt>charbuffer</tt> type.
@param value Value to convert. The value must be of type <tt>int</tt>, <tt>float</tt>,
<tt>double</tt>, <tt>buffer</tt>, <tt>charbuffer</tt>, <tt>string</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>charbuffer</tt>
@note <ul>
<li>For numeric types (<tt>int</tt>, <tt>float</tt> and <tt>double</tt>) returned charbuffer
contains a string representation of argument. Floating point values are converted
using %g format specification.</li>
<li>In case of a <tt>buffer</tt> argument all bytes of argument are interpreted as 
single character values, independent of current character size. </li>
<li>In case of a <tt>charbuffer</tt> function simply returns a reference (not a copy!) to argument.</li>
<li>Values of <tt>object</tt> type must have a <tt>charbuffer</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
charbuffer charbuffer(var value);

/** Convert value to <tt>vector</tt>.
Function tries to convert argument into value of <tt>vector</tt> type.
@param value Value to convert. The value must be of type <tt>vector</tt>, 
<tt>buffer</tt>, <tt>charbuffer</tt>, <tt>dictionary</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>vector</tt>
@note <ul>
<li>In case of a <tt>vector</tt> function simply returns a reference (not a copy!) of argument.</li>
<li>In case of a <tt>buffer</tt> argument each byte of argument is converted into vector
element if type <tt>int</tt>. </li>
<li>In case of a <tt>charbuffer</tt> argument each character of argument is converted into vector
element if type <tt>int</tt>. </li>
<li>In case of a <tt>dictionary</tt> argument result vector consists of values from dictionary. </li>
<li>Values of <tt>object</tt> type must have a <tt>vector</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
vector vector(var value);

/** Convert value to <tt>dictionary</tt>.
Function tries to convert argument into value of <tt>dictionary</tt> type.
@param value Value to convert. The value must be of type <tt>vector</tt>, 
 <tt>dictionary</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>vector</tt>
@note <ul>
<li>In case of a <tt>vector</tt> result contains vector elements as values where keys
are string representations of element indexes in vector.</li>
<li>In case of a <tt>dictionary</tt> function simply returns a reference (not a copy!) of argument.</li>
<li>Values of <tt>object</tt> type must have a <tt>dictionary</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
dictionary dictionary(var value);

/** Get type id of argument value.
@param value value to get type id from
@return type id of <i>value</i>
@note Use #gettypename function to convert returned id into readable string representation.
@see gettypename, BppValueType
*/
int gettype(var value);


/** Get human readable name of a data type.
@param typeid id of data type to get name of
@param prefixed Specifies if result is prefixed by 'BVT_' in order to get
internal symbol name of type. Prefix is used if \e prefixed can be evaluated to \c true.
@return Name of data type. Following table shows possible names:
<table border="1">
<tr><td><b>Type id (HEX)</b></td><td><b>Type name</b></td><td><b>Type</b></td></tr>
<tr><td>0 (00)</td><td>NULL</td><td><tt>null</tt></td></tr>
<tr><td>1 (01)</td><td>VOID</td><td>void</td></tr>
<tr><td>2<(02)/td><td>AUTO</td><td>auto</td></tr>
<tr><td>64 (40)</td><td>SBYTE</td><td><tt>sbyte</tt></td></tr>
<tr><td>65 (41)</td><td>BYTE</td><td><tt>byte</tt></td></tr>
<tr><td>68 (44)</td><td>SHORT</td><td><tt>short</tt></td></tr>
<tr><td>69 (45)</td><td>USHORT</td><td><tt>ushort</tt></td></tr>
<tr><td>72 (48)</td><td>INT</td><td><tt>int</tt></td></tr>
<tr><td>73 (49)</td><td>UINT</td><td><tt>uint</tt></td></tr>
<tr><td>74 (4A)</td><td>INTPTR</td><td><tt>intptr</tt></td></tr>
<tr><td>76 (4C)</td><td>LONG</td><td><tt>long</tt></td></tr>
<tr><td>77 (4D)</td><td>ULONG</td><td><tt>ulong</tt></td></tr>
<tr><td>104 (68)</td><td>FLOAT</td><td><tt>float</tt></td></tr>
<tr><td>108 (6C)</td><td>DOUBLE</td><td><tt>double</tt></td></tr>
<tr><td>128 (80)</td><td>FILE</td><td><tt>FILE</tt></td></tr>
<tr><td>129 (81)</td><td>CODE</td><td>native function</td></tr>
<tr><td>130 (82)</td><td>VAR</td><td>variable, dictionary entry</td></tr>
<tr><td>160 (A0)</td><td>OBJECT</td><td><tt>object</tt></td></tr>
<tr><td>161 (A1)</td><td>STRONGOBJECT</td><td><tt>strong typed object</tt></td></tr>
<tr><td>168 (A8)</td><td>CLASS</td><td><tt>class</tt></td></tr>
<tr><td>176 (B0)</td><td>STRING</td><td><tt>string</tt></td></tr>
<tr><td>177 (B1)</td><td>BUFFER</td><td><tt>buffer</tt></td></tr>
<tr><td>178 (B2)</td><td>CHARBUFFER</td><td><tt>charbuffer</tt></td></tr>
<tr><td>180 (B4)</td><td>VECTOR</td><td><tt>vector</tt></td></tr>
<tr><td>181 (B5)</td><td>BYTECODE</td><td>B++ function</td></tr>
<tr><td>182 (B6)</td><td>LOCALVERINFO</td><td>local variable info (internal used)</td></tr>
<tr><td>184 (B8)</td><td>DICTIONARY</td><td><tt>dictionary</tt></td></tr>
<tr><td>224 (E0)</td><td>OBJECTREF</td><td>free reference to <tt>object</tt></td></tr>
<tr><td>232 (E8)</td><td>CLASSREF</td><td>free reference to <tt>class</tt></td></tr>
<tr><td>240 (F0)</td><td>STRINGREF</td><td>free reference to <tt>string</tt></td></tr>
<tr><td>241 (F1)</td><td>BUFFERREF</td><td>free reference to <tt>buffer</tt></td></tr>
<tr><td>242 (F2)</td><td>CHARBUFFERREF</td><td>free reference to <tt>charbuffer</tt></td></tr>
<tr><td>244 (F4)</td><td>VECTORREF</td><td>free reference to <tt>vector</tt></td></tr>
<tr><td>245 (F5)</td><td>VARINFOREF</td><td>free reference to B++ function</td></tr>
<tr><td>246 (F6)</td><td>CODEREF</td><td>free reference to native function</td></tr>
<tr><td>248 (F8)</td><td>DICTIONARYREF</td><td>free reference to <tt>dictionary</tt></td></tr>
</table>
@note Predefined literals of symbolc type names should be used for <i>typeid</i> value rather than
numeric constants, e.g. BVT_INT instead of 72.
@see gettype
*/
string gettypename(int typeid, bool prefixed = false);

/** Get class name of object or class.
@param value <tt>object</tt> or <tt>class</tt> to get class name from
@return If <i>value</i> is of type <tt>object</tt> or <tt>class</tt> function returns
class name, otherwise <tt>null</tt>.
@note Because return value is only a reference (not a copy) of class name stored in BppClass instance,
client code should not change contents of return value.
@see dynamic_cast, gettype
*/
string getclassname(var value);

/** Gets a global symbol (variable, function or class) by name at runtime.
@param name Name of symbol to search for.
@param isclass If set to <tt>true</tt> function searches for a class, otherwise for
global symbol of any other type.
@return If a symbol for <i>name</i> was found, function returns its value, e.g. function
pointer for a function, otherwise <tt>null</tt>.
@note This function is mainly intended for using different system functions, depending
on platform used, in same program without compiler error.
@par Example
@code
// Gets current working directory
string getCwd()
{
   string buf = null;
   if (MSDOS)
   {
      buf = newstring(256);
      function segread = getsymbol("segread");
      function setreg = getsymbol("setreg");
      function int86x = getsymbol("int86x");
      segread();
      setreg("ah",0x47);
      setreg("dl",0);
      intptr pbuf = addr(buf);
      setreg("ds",pbuf >> 16);
      setreg("si",pbuf & 0xFFFF);
      int86x(0x21);
      return buf;
   }
   else if (WIN32)
   {
      function GetFunc = getsymbol("GetKernel32Func");
      intptr pfunc = GetFunc("GetCurrentDirectory?");
      uint siz = CallLibFunc(pfunc,0,0);
      buf = newstring(siz);
      CallLibFunc(pfunc,siz,addr(buf));
      return buf;
   }
   throw "Invalid OS";
}
@endcode
@see dynamic_cast, gettype
*/
var getsymbol(string name, bool isclass = false);

/** Do dynamic cast to specified destination class.
Function behaves very similarly to <tt>dynamic_cast</tt> operator in C++.<br>
First it evaluates expression <i>expr</i>. If evaluation results in an object of class
type <i>desttype</i> or derived class, it returns the object, otherwise <tt>null</tt>.
@param desttype class to cast <i>expr</i> to
@param expr expression of any type
@return object of class <i>desttype</i> (or derived) on success, otherwise <tt>null</tt>
@note Other than <tt>dynamic_cast</tt> operator in C++ there is no exeption thrown
if <i>expr</i> is not an object. In this case function simply returns <tt>null</tt>.
Use #gettype to check a value for its basic type.
@par Example:
@code
   class Base
   {
      Base();
   }
   Base::Base() { }

   class Derived : Base
   {
      Derived();
   }
   Derived::Derived() { }

   main(;obj1, obj1, o, i)
   {
      obj1 = new Base();
      obj2 = new Derived();
      o = dynamic_cast(Base, obj1);      // returns obj1
      o = dynamic_cast(Base, obj2);      // returns obj2
      o = dynamic_cast(Derived, obj1);   // returns null
      o = dynamic_cast(Derived, obj2);   // returns obj2
      i = dynamic_cast(Base, 17);        // returns null (not 0)
   }
@endcode
@see gettype, dynamic_cast(string,var)
*/
object dynamic_cast(class desttype, var expr);

/** Do dynamic cast to destination class specified by name.
Function behaves very similarly to <tt>dynamic_cast</tt> operator in C++.<br>
First it evaluates expression <i>expr</i>. If evaluation results in an object of class
specified by <i>classname</i> or derived class, it returns the object, otherwise <tt>null</tt>.
@param classname name of class to cast <i>expr</i> to
@param expr expression of any type
@return object of class <i>desttype</i> (or derived) on success, otherwise <tt>null</tt>
@note Other than <tt>dynamic_cast</tt> operator in C++ there is no exeption thrown
if <i>expr</i> is not an object. In this case function simply returns <tt>null</tt>.
Use #gettype to check a value for its basic type.
@par Example:
@code
   class Base
   {
      Base();
   }
   Base::Base() { }

   class Derived : Base
   {
      Derived();
   }
   Derived::Derived() { }

   main(;obj1, obj1, o, i)
   {
      obj1 = new Base();
      obj2 = new Derived();
      o = dynamic_cast("Base", obj1);               // returns obj1
      o = dynamic_cast("Base", obj2);               // returns obj2
      o = dynamic_cast("Derived", obj1);            // returns null
      o = dynamic_cast("Derived", obj2);            // returns obj2
      o = dynamic_cast(getclassname(obj2), obj1);   // returns null
      o = dynamic_cast(getclassname(obj1), obj2);   // returns obj2
      i = dynamic_cast("Base", 17);                 // returns null (not 0)
   }
@endcode
@see gettype, getclassname, dynamic_cast(class, var)
*/
object dynamic_cast(string classname, var expr);

/// @}

/** @addtogroup TimeFuncs Date and time functions
@{
*/
/** Get C-timestamp.
@return number of seconds elapsed since 1970/01/01 0:00 GMT
@note Function behaves like <tt>time</tt> function of C standard library, on 16-Bit-platforms return type is <i>int</i> (32 bits)
instead of <i>long</i>.
*/
long time();

/** Get elapsed time.
This function is intended to be used for time measurements.<br>
The behavior of function differs depending on platform. On Windows platforms
(including Windows CE) it returns elapsed time since start of system where
for DOS platforms elapsed time since start of current program is returned.
@return elapsed time in milliseconds
@note Depending on platform, resolution of timer may be lower than 1 millisecond.
@par Example:
@code
   main()
   {
      uint t1 = timer();
      // do anything here
      ...
      uint t2 = timer();
      print(t2-t1, " ms needed for operation.\n");
   }
@endcode
*/
uint timer();


/** Get string representation of a C-timestamp.
Function makes a readable string from timestamp. It behaves like same named
C function.
@param t C-timestamp, e.g. result of a #time() call. On 16-Bit-platforms argument type is <i>int</i> (32 bits).
@return string representation of argument, e.g. <i>Sat May 10 21:52:54 2008</i>
*/
string ctime(long t);

/** Make a localized time point from C-timestamp and split it into
different parts.
Behavior of function is similarly to same named C function. However,
because B++ doesn't support structure types it uses a vector for result.
@param t Timestamp to localize, on 16-Bit-platforms argument type is <i>int</i> (32 bits).
@return vector of following components:
<table border="1">
   <tr><td align="center"><b>Index</b></td><td align="center"><b>Value</b></td></tr>
   <tr><td>0</td><td>Seconds after minute (0 .. 59).</td></tr>
   <tr><td>1</td><td>Minutes after hour (0 .. 59).</td></tr>
   <tr><td>2</td><td>Hours after midnight (0 .. 23).</td></tr>
   <tr><td>3</td><td>Day of month (1 .. 31).</td></tr>
   <tr><td>4</td><td>Month (0 .. 11; January = 0).</td></tr>
   <tr><td>5</td><td>Year (current year minus 1900).</td></tr>
   <tr><td>6</td><td>Day of week (0 .. 6; Sunday = 0).</td></tr>
   <tr><td>7</td><td>Day of year (0 .. 365; January 1 = 0), -1 if not supported.</td></tr>
   <tr><td>8</td><td>Positive value if daylight saving time is in effect;<br>
   0 if daylight saving time is not in effect;<br>negative value if status is unknown.</td></tr>
</table>
@note On Windows CE informations about day of year (index 7) and daylight saving time
are currently not supported, i.e. both values will be -1.
see time, ctime
*/
vector localtime(long t);

/// @}

/** @addtogroup RandFuncs Random number functions
@{
*/
/** Get random number.
Function computes a pseudo-random number. It behaves like same named C function.
@return random number between 0 and 0x7FFF
@note Multiple calls to this function will produce an approximately uniformly distributed
series of random numbers. However, random number generator should be initialized
using #srand function to prevent from generating same series for each run of your program.
@see srand
*/
int rand();

/** Initialize random number generator.
Function initializes random number generator by value of <i>seed</i>.
It behaves like same named C function.
@param seed value to initialize random number generator with
@return null
@note On start of program random number generator is initialized by value 1.
In order to get "really random" series of random numbers, usually a time depended
value for <i>seed</i> is used in calls to this function.
@par Random number generator example:
@code
   main(;i,j,l,seed)
   {
      // create 5 different series of random numbers
      for (l = 0; l< 5; ++l)
      {
         seed = timer(); // use elapsed time for initialization
         srand(seed);    // initialize generator
         print("initial value; ",seed,"\n");
         // create series of 100 numbers and print
         // 10 numbers per line
         for (i=0; i<10; ++i)
         {
            for (j=0; j<10; ++j) print(rand()," ");
            print("\n");
         }
      }
   }
@endcode
@see rand, timer, print
*/
null srand(int seed);

/// @}

/** @addtogroup ProgFuncs Program related functions
@{
*/

/** Get command line arguments passed to B++.
Function builds a vector of all command line arguments passed to B++ on run.
When B++ command line version is running, first argument is always the program
name itself (e.g. bp2.exe).
@return Vector of all command line arguments. All elements of returned
vector are of <tt>string</tt> type.
@note When B++ is used as a library, i.e. inside any program, owner program is resposnible
for passing arguments to B++ if needed.
@see getusrargs, BppVirtualMachine::execute
*/
vector getargs();

/** Get command line user arguments passed to B++.
Function builds a vector of all command line user arguments passed to B++ on run.
First element of result vector is first command line argument behind hash sign(#)
in command line.
@return Vector of all command line user arguments. All elements of returned
vector are of <tt>string</tt> type.
@note When B++ is used as a library, i.e. inside any program, owner program is resposnible
for passing arguments to B++ if needed.
@par Example for using getargs and getusrargs functions:
@code
   main(;vec1,vec2,n,i)
   {
      vec1 = getargs();    // get all arguments
      vec2 = getusrargs(); // get user arguments
      print("args:\n");
      n = vecsize(vec1);
      for (i=0;i<n;++i) print(vec1[i],"\n");
      print("usrargs:\n");
      n = vecsize(vec2);
      for (i=0;i<n;++i) print(vec2[i],"\n");
   }
@endcode
@see getargs, BppVirtualMachine::execute
*/
vector getusrargs();

/** Get number of function arguments.
Function retrieves number of arguments passed to calling function.
@return number of arguments of caller
@note If called from <tt>main</tt> in a command line program function of a program, all command line arguments (including programm name)
are passes as string parameters to function, i.e. arguments are same as result of #getarg function.
@see arg, getargs getusrargs
*/
int argcnt();

/** Get function argument of caller by index.
Function returns argument of specified index from calling function. It may
be used in conjunction with #argcnt to construct functions of variable sized
argument lists.
@param idx index of argument to get
@return callers function argument for index <i>idx</i>
@note In case of invalid <i>idx</i> value an error is reported.
@par Example for using arg and argcnt functions
@code
   // define a function
   var f(var x)   // only first argument has got a name
   {
      uint n = argcnt();            // get number of arguments passed
      print("\n",n," arguments:\n");
      for (uint i=0;i<n;++i)
         print("\t",arg(i));   // output argument values
      print("\n---\n");
      return x; // return value of named (first) argument
   }

   main()
   {
      print(f(1,3,"test"));
      // should print
      // 3 arguments:
      //    1  3  test
      // ---
      // 1
   }
@endcode
@see argcnt
*/
var arg(int idx);

/** Gets vector of functions actual parameters.
Function returns vector of all arguments passed to calling function. It may
be used in conjunction with #call function in order to pass arguments
to another function.
@return Vector containing all actual parameters  of caller. If #argv is called
from an element function, first element of result is the calling object.
@see argcnt, arg
*/
vector argv();

/** Calls function <i>f</i> with actual parameters from <i>args</i>.
Function does a generic call to another function passing elements of <i>args</i>
as actual parameters to called function. It is mainly intended to dispatch
variable number of arguments to any function.
@param f Function or functor object to be invoked
@param args Vector of arguments to pass to <i>f</i>.
@return Returns return value of <i>f</i> if any, otherwise result is undefined.
@note Function may be used for call to element functions of objects too. In that case
first element of <i>args</i> must be the object itself.<br>
If <i>f</i> is an object, it must define <tt>operator()</tt> (or element function <tt>OP_CALL</tt>).
@par Examples for using call function
<b>Building a C-like printf function:
@code
int printf(string format)
{
   vector v = argv();
   print(call(newstring,v));
   return int(size(v))-1;
}
@endcode
<b>Simple implementation of a .NET-like delegate
@code
class Delegate
{
private:
   object _obj;
   function _func;
public:
   Delegate(object obj, function func)
   {
      _obj = obj;
      _func = func;
   }
   operator ()() {
     vector v = argv();
     if (_obj) v[0] = _obj; // Object is first parameter for call.
     else verase(v,0,1);
     return call(_func,v);
  }
}
@endcode
*/
var call(var f, vector args = null);

/** Compile source code at runtime.
Function compiles source code at runtime and integrates it to running program.
@param src string containing B++ source code to compile
@param srcname Name of source. If specified <i>srcname</i> is included in
debug output in case of an error during compiling.
@param debug If <i>true</i>, generated bytecode contains additional information about local
variable names and line nmbers in source code for debugging purposes.
@return <tt>true</tt> (1) if <i>src</i> was compiled successfully, otherwise <tt>false</tt> (0)
@note <ul>
<li>Other than <tt>\#include</tt> processing instructon code is compiled and integrated
at runtime.</li>
<li>This function may be called anywhere in your program. However, result is mapped into
globally context replacing already existing symbols of same names (if any) by new
versions. That allows modifying behavior of running programm (e.g. depending on
any conditions).</li>
</ul>
@par Example:
@code
   // define a test function
   testfunc(n) {
      print("testfunc(\"",n,"\") called.\n");
   }

   main(;fstr) {
      // call testfunc
      testfunc("Hello World");
      // create source of alternate function
      fstr = 
		  R"(testfunc(n) {
			   print("alternate version of testfunc called:",n,"\n");
			 })";
      if (compile(fstr))
         testfunc("Hello World");
   }
@endcode
@see compile(FILE,string), loadmodule, compilefn
*/
bool compile(string src, string srcname = "", bool debug = false);

/** Compile source file at runtime.
Function compiles source file at runtime and integrates it to running program.
@param fp file opened for binary read
@param srcname Name of source. If specified <i>srcname</i> is included in
debug output in case of an error during compiling.
@param debug If <i>true</i>, generated bytecode contains additional information about local 
variable names and line nmbers in source code for debugging purposes.
@return <tt>true</tt> (1) if contents of <i>fp</i> was compiled successfully, otherwise <tt>false</tt> (0)
@note <ul>
<li>Other than for <tt>\#include</tt> processing instructon code is compiled and integrated
at runtime.</li>
<li>This function may be called anywhere in your program. However, result is mapped into
globally context replacing already existing symbols of same names (if any) by new
versions. That allows modifying behavior of running programm (e.g. depending on
any conditions).</li>
</ul>
@see compile(string,string), loadmodule, compilefn
*/
bool compile(FILE fp, string srcname = "", bool debug = false);

/** Compiles source of a single function or element function at runtime.
Function compiles source code of a function body at runtime without integration into running program.
@param src Source code of function body to compile.
@param args Contents of function argument list in same notation as for a 'normal' function decvlaration.
@param rtype Name of return type of function. Empty value means no specific type.
@param ownerClass Type or class name of owner class, i.e. class to create element function for. If
not specified, a simple function is created.
@param srcname Name of source. If specified <i>srcname</i> is included in
debug output in case of an error during compiling.
@param debug If <i>true</i>, generated bytecode contains additional information about local
variable names and line nmbers in source code for debugging purposes.
@return Compiled function (or element function) if contents of <i>src</i> was compiled successfully, otherwise <tt>null</tt>.
@note <ul>
<li>Function allows dynamic craetion of any function at runtime and assign it e.g. to a variable or call it directly without 
sideeffects for code of running program.</li>
<li>Inside body (contents of <i>src</i>) all globally defined symbols (but not local symbols of caller) and
members of ownerclass (if specified by <i>ownerClass</i>) can be accessed.</li>
<li>If a member function is created, result must be called in context of an object of 
<i>ownerClass</i> type.</li>
</ul>
@par Example
@code
   // Evaluate expressions at runtime
   var evalExpr(string expr)
   {
      var f = compilefn("return ("+expr+");");
      if (f) return f();
      return null;
   }
   void main()
   {
      while (true)
      {
	     print("Expression: ");
	     string s = gets();
		 if (size(s) == 0) break;
		 print(" Result: ",evalExpr(s),"\n")
	  }
   }
@endcode
 compile(string,string), loadmodule
*/
function compilefn(string src, string args = "", string rtype = "", var ownerClass = null, string srcname = "", bool debug = false);

/** Parses a literally expression or single literally value from string.
@param src Source string to parse
@param srcname Source name for use in error messages
@return Result of expression in <i>src</i>
@note If parsing fails, i.e. in case of invalid expression, exception of type <tt>string</tt> containing error description is thrown.
*/
var litexpr(string src, string srcname= "");

/** Load module at runtime.
Function loads a pre-compiled byte code module at runtime and integrates it to running program.
@param filename Name of file containing module
@return 1 on success, otherwise 0
@note Other than for <tt>\#use</tt> processing instructon code is integrated
at runtime.<br>
This function may be called anywhere in your program. However, result is mapped into
globally context replacing already existing symbols of same names (if any) by new
versions. That allows modifying behavior of running programm (e.g. depending on
any conditions).
@see compile(string,string), compile(FILE, string)
*/
int loadmodule(string filename);

/// @}

/** @addtogroup SysFuncs System functions
@{
*/

/** Get address of value data.
Function returns address of data for value <i>val</i>. It is intended for
use with system specific functions which require variable pointers for arguments.
@param any value or variable of any type except <tt>null</tt>
@return pointer representing value address. The returned value depends on
data type of <i>any</i> as follows:
<ul>
  <li>For types <tt>string</tt>, <tt>buffer</tt> and <tt>charbuffer</tt>
  function returns start address of data area. The result can be interpreted
  as an array or string in C-style.</li>
  <li>For all numeric types current function does same as explicit conversion into <tt>intptr</tt> type.</li>
  <li>In case of any other type, value itself is returned as a pointer.
  This allows passing of reference data (e.g. file pointer or object) to
  native functions.</li>
</ul>
*/
intptr addr(var any);

/** Does raw copy of specified block in memory.
Function behaves like same named C function, i.e. it copies <i>n</i> bytes
from <i>src</i> to <i>dst</i>, where overlapping of memory areas is allowed.
@param dst Destination address of copy operation
@param src Source address of copy operation
@param n Number of bytes to copy
@return Destination address (same as <i>dst</i>)
@note
<ul>
<li>Function does implicit cast of argument types.</li>
<li>On DOS platforms function allows copying blocks larger than 64kB too.</li>
<li>In case of invalid arguments or an error while copy operation, it 
throws exception of <tt>string</tt> type.</li>
<li><b>Caution:</b> Function allows manipulation of any memory area and
may crash system!</li>
</ul>
*/
intptr memmove(intptr dst, intptr src, uint n);

/** Runs a command line as a synchronous child process.
Function invokes command line interpreter with specified command line and returns execution result of call.
@param cmd Command line passed to system. Line may contain name (and optionally path) of programm to run
or name of internal command followed by any parameters.
@return Execution result value of command.
@note Function is not available in Windows-CE builds. Furthermore call fails, if there is no command processor specified
by COMSPEC environment variable of system (e.g. on DIP-DOS of Atari Portfolio).<br>
If call fails, function throws exception.
*/
int system(string cmd);

/** Sets or gets current runtime localization scheme.
Function behaves like same named C function (and internally uses that function).<br>
@param category Category to set/get localization parameters for. Valid values are
<ul>
<li><tt>LC_ALL</tt> (0): all categories</li>
<li><tt>LC_COLLATE</tt> (1): string comparison</li>
<li><tt>LC_CTYPE</tt> (2): character fuctions</li>
<li><tt>LC_MONETARY</tt> (3): currency format</li>
<li><tt>LC_NUMERIC</tt> (4): numeric formats</li>
<li><tt>LC_TIME</tt> (5): time format</li>
</ul>
@param locale Locale string, e.g. "en-US", "German_Germany.437", ".OCP" or "C".<br>
Empty string means 'System default settings', <tt>null</tt> returns current settings
without any changes.
@return String of settings after the call or emty string in case of an error.
@note Function (especially valid values for <i>locale</i> string) depends on Operating System.<br>
On DOS-Platforms it is useless, because only "C"-locale is available.
*/
string setlocale(int category, string locale);


/** @name Environment access functions 
@note <p>For Windows CE builds environment is emulated in registry
(key HKLM\Environment). There is only one globally defined environment, i.e. 
modification of an environment variable may affect other processes too.</p>
<p>In DOS and Windows Desktop builds process local environment area is used.</p>
*/
//@{

/** Gets dictionary containing copy of environment variables.
@return Ordered dictionary containing copies of a environment variable values.
Keys are the names of the environment variables.
@par Example
@code
   // list names and values of all environment variables:
   void main()
   {
      auto env = getenv();
      vector names = dictgetkeys(env);
      uint n = size(names);
      for (uint i=0;i<n;++i) print(names[i],"=",env[names[i]],"\n");
   }
@endcode
@note Function won't work for Windows CE (always returns empty dictionary).
@see getenv(string), setenv
*/
dictionary getenv();

/** Get value of specified environment variable.
Function returns value of environment variable <i>varname</i>.
@param varname Name of environment variable to get value from.
@return Copy of environment variable value if it exists or
empty string otherwise.
@par Example
@code
   // list all parts of "Path" environment variable:
   void main()
   {
      string path = getenv("Path");
      vector parts = strsplit(path,";");
      uint n = size(parts);
      for (uint i=0;i<n;++i) print(parts[i],"\n");
   }
@endcode
@see getenv(), setenv
*/
string getenv(string varname);

/** Set value of an environment variable.
Function defines a new environment variable or changes value of an existing one.
@param varname Name of environment variable so define or change.
@param value Value to set for environment variable. If specified value is an empty string,
variable is removed from environment.
@return 1 on success, otherwise 0
@see getenv
*/
int setenv(string varname, string value);

/** Gets code of last system error.
Function reads errno value of runtime.
@return Code of last error occurred while a runtime function call.
@note Function may be used to get error information in case of failure of a call to function in
runtime library, e.g. an I/O-function.
@see strerror(), errno(int)
*/
int errno();

/** Sets code of last system error.
Function sets errno value of runtime and returns changed value.
@param New valus of system error code
@return Value of system error code after setting (should be same as <i>err</i>).
@note Function may be used to reset pending system error (e.g. after a numerical overflow)
@see errno()
*/
int errno(int err);

/** Gets number of system error codes.
@return Number of error codes defined by system.
@note Function may be used to validate an error code or enumerate system error messages.
@par Example for enumerate system error codes:
@code
void main()
{
    int n = nerr();
    for (int i=0;i<n;++i)
       print(i,"\t",strerror(i),"\n");
}
@endcode
@see strerror(int errcode)
*/
int nerr();

/** Gets last system error message.
Function calls errno() and returns error message accociated to its result.
@return Description of last error occurred while a runtime function call.
@note Function may be used to get error information in case of failure of a call to function in
runtime library, e.g. an I/O-function.
@see errno(), strerror(int errcode)
*/
string strerror();

/** Gets error message for specified code.
Function returns error message accociated to <i>errcode</i>.
@param errcode Code of error to get description for
@return Description of specified error code. On invalid coed it returns "Unknown error".
@see errno(), strerror()
*/
string strerror(int errcode);

//@}

/** @name Dynamic library (DLL) functions.
@note For DOS builds handling of Dynamic Link Libraries (DLLs) is
supported using Kiez<i>Soft</i> DOSDLL library. Please refer to
documentation of this library for restrictions and detailed using
instructions.
*/
//@{

/** Load a dynamic library.
Function loads dynamic library specified by <i>dllname</i> into current process
and returns a handle to it.
@param dllname Name of library. The name specified is the file name of library module (including extension if any).
It may contain a relative or absolute path specification.
@return handle to loaded library or 0 (if not found)
@note <ul><li>The returned handle is needed as argument for following calls to
<tt>GetProcAddress</tt> and <tt>FreeLibrary</tt> functions.</li>
<li>A library loaded by <tt>LoadLibrary</tt> function should be released using
<tt>FreeLibrary</tt> if no longer needed.</li>
<li>If no absolute full path is specified in <i>dllname</i>, a system dependend serach for 
library is done. Please refer to platform documentation (or DOSDLL manual for DOS builds)
for search strategy used.</li>
<li>If <i>dllname</i> contains an asterisk (*), the asterisk character is replaced by build specific name 
extension, consisting of following (optional) parts:
<ul>
   <li>u : marks UTF-16 build</li>
   <li>d : marks Debug build</li>
   <li>_64 : marks build for 64 Bits platform</li>
</ul>
</li>
<li>In DOS builds current program is aborted if library coudn't be loaded.</li>
</ul>
@par Example for using dynamic library functions:
@code
   // This sample program works for Win32 Desktop platforms.
   // For Windows CE load coredll.dll instead of user32.dll .
   #define MB_ICONINFORMATION 0x40
   // function shows a simple message box
   ShowMessage(string title, string msg)
   {
      intptr hinst = LoadLibrary("user32.dll");
      // ? will be replaced by 'W' or 'A', depending if unicode build or not
      intptr fptr = GetProcAddress(hinst,"MessageBox?");
      // use addr function to get C-character pointers from strings
      CallLibFunc(fptr,0,addr(msg),addr(title),MB_ICONINFORMATION);
      FreeLibrary(hinst);
   }

   main()
   {
      ShowMessage("Hello World","This is a message\nfrom B++.");
   }
@endcode
@see FreeLibrary, GetProcAddress, CallLibFunc, addr
*/
intptr LoadLibrary(string dllname);

/** Release dynamic library.
Function releases a dynamic library previously loaded by LoadLibrary function.
@param hdll Handle of library to release.
@return 1 on success, otherwise (i.e. on invalid handle) 0
@see LoadLibrary
*/
int FreeLibrary(intptr hdll);

/** Get address of function or data area exported from dynamic library.
Function returns address of symbol <i>name</i> (function or data area) exported by
a dynamic library.
@param hdll library handle
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned in
   Windows builds. In DOS builds program is aborted in that case.</li>
   <li>In Windows builds <i>name</i> may contain a question sign (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
</ul>
@see LoadLibrary, CallLibFunc
*/
intptr GetProcAddress(intptr hdll, string name);

/** Call a function inside dynamic library.
Function invokes library function at address specified by <i>procaddr</i>.
@param procaddr Adress of function to invoke. Use #GetProcAddress to get address from
function name.
@param ... Comma separated list of parameters needed by called function.
@return result of function call
@note<ul>
   <li>Because B++ knows nothing about function called inside the library, there is
   no way to do any checks. So programmer is responsible to take care about
   passing correct parameters.</li>
   <li>All parameters must be passed as values of same width as <tt>intptr</tt> type,  values
   convertible to <tt>intptr</tt> may be passed directly where for all other types addresses should be used.</li>
   <li>Functions inside dynamic library must be implemented using C or STDCALL calling conventions.
   Furthermore they should return a result value of <tt>intptr</tt> size.</li>
   <li>Meaning of returned value depends on implemantation of library function. If library
   function returns nothing, result is undefined.</li>
   <li>In Windows CE and on x64 platforms number of function arguments is restricted to a maximum of 16.</li>
</ul>
@par Example:
In the following we assume a dynamic library "database.dll" implementing some database access functions.
This library exports a login function of following C-signature:
@code
   int login(const char* userName, const char* passwd);
@endcode
This function needs two parameters (user name and password). On success it returns a session id for database access,
otherwise 0.<br>
Sample program below shows all steps for call the function.
@code
   // Example for DLL function call
   void main()
   {
      // first load dynamic library
      intptr hdll = LoadLibrary("database.dll");
      // get address of exported function
      intptr proc = GetProcAddress(hdll,"login");
      // now ask user for her/his name and password
      print("Enter your name: ");
      string name = gets(); 
      print("Enter your password: ");
      string pwd = gets();
      // try to login
      // dont forget to pass addresses of strings
      intptr session = CallLibFunc(proc,addr(name),addr(pwd));
      if (session)
      {
         // do anything
      }
      else
         print("Database access denied!\n");
      // release library
      FreeLibrary(hdll); 
   }
@endcode
@see LoadLibrary, GetProcAddress, FreeLibrary, addr
*/
intptr CallLibFunc(intptr procaddr, ...);
//@}

//@}
/** @addtogroup WINFuncs Windows specific system functions.
All functions of this section are available in Windows builds only.
@{
*/
/** Get address of function or data area exported from user32.dll library.
Function returns address of symbol <i>name</i> (function or data area) exported by
the library. It is intended to simplify using of Windows-API functions.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
   <li>In WindowsCE builds symbol is retrieved from coredll.dll library.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress, GetCoreFunc
*/
intptr GetUser32Func(string name);

/** Get address of function or data area exported from shell32.dll library.
Function returns address of symbol <i>name</i> (function or data area) exported by
the library. It is intended to simplify using of Windows-API functions.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
   <li>In WindowsCE builds symbol is retrieved from coredll.dll library.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress, GetCoreFunc
*/
intptr GetShell32Func(string name);

/** Get address of function or data area exported from gdi32.dll library.
Function returns address of symbol <i>name</i> (function or data area) exported by
the library. It is intended to simplify using of Windows-API functions.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
   <li>In WindowsCE builds symbol is retrieved from coredll.dll library.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress, GetCoreFunc
*/
intptr GetGdi32Func(string name);

/** Get address of function or data area exported from kernel32.dll library.
Function returns address of symbol <i>name</i> (function or data area) exported by
the library. It is intended to simplify using of Windows-API functions.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
   <li>In WindowsCE builds symbol is retrieved from coredll.dll library.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress, GetCoreFunc
*/
intptr GetKernel32Func(string name);

/** Get address of function or data area exported from ole32.dll library.
Function returns address of symbol <i>name</i> (function or data area) exported by
the library. It is intended to simplify using of Windows-API functions.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress
*/
intptr GetOle32Func(string name);

/** Alias of GetUser32Func.
Because in most functions of WindowsCE-API are contained in coredll.dll library,
there are no user32, shell32, kernel32 and gdi32 libraries in WindowsCE.<br>
This function alias is intended to be used in programms running on WinCE only.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
   <li>In WindowsCE builds symbol is retrieved from coredll.dll library.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress, GetUser32Func
*/
intptr GetCoreFunc(string name);

/** Get last error code.
Function simply wraps Windows API function of same name.
@return Last error code of calling thread.
@see GetSysErrorMsg
*/
uint GetLastError();

/** Get readable message for system error code.
Function retrieves error message for last system error code.
@return Error message for last system error code, empty string if no message available.
@note GetLastError is called internally to get error code.
@see GetLastError
*/
string GetSysErrorMsg();

/** Get readable message for system error code.
Function retrieves error message for specified code.
@param errcode Error code to get message for (optional)
@return Error message for specified code, empty string if no message available.
@see GetLastError
*/
string GetSysErrorMsg(uint errcode);

/// @}
/** @addtogroup DOSFuncs DOS specific system functions.
All functions of this section are available in DOS builds only.
@{
*/



/** Get size of available memory. 
Function returns amount of free memory on DOS heap.
@return number of free bytes in memory
*/
int memsize();

/** Set register value.
Function sets value of register specified by <i>regname</i>.
@param regname Name of register to set; possible names are:<br>
<tt>ax, bx, cx, dx, di, si, al, ah, bl, bh, cl, ch, dl, dh,
   cflag, flags, ds, es, ss, cs</tt>.
@param value New value of specified register. For 8-bit registers
lowest byte and for 16-bit registers low word of <i>value</i> is used.
@return null
@note <ul>
   <li>Register names in <i>regname</i> argument are not case sensitive, i.e. 
   "AX" means same as "ax".</li>
   <li>Function stores values in internal register structure rather immediately
   modifying processor registers. Contents of this structure is applied to
   processor registers on calls to #int86 and #int86x functions.</li>
</ul>
@see getreg, segread, int86, int86x
*/
null setreg(string regname, int value);

/** Get register value.
Function gets value of register specified by <i>regname</i>.
@param regname Name of register to get value from; possible names are:<br>
<tt>ax, bx, cx, dx, di, si, al, ah, bl, bh, cl, ch, dl, dh,
cflag, flags, ds, es, ss, cs</tt>.
@return value of specified register
@note <ul>
<li>Register names in <i>regname</i> argument are not case sensitive, i.e. 
"AX" means same as "ax".</li>
<li>Function reads values from internal register structure rather directly reading 
processor registers. Contents of this structure is modified explicitely using
functions #setreg or #segread and implicitely by #int86 and #int86x
functions.</li>
</ul>
@see setreg, segread, int86, int86x
*/
int getreg(string regname);

/** Read segment registers.
Function reads current values of processors segment registers (DS, ES, SS, CS)
and stores them into internal register data structure.
@return null
@note Use #getreg to access values from your program.
@see getreg, setreg
*/
null segread();

/** Do x86 interrupt call.
Function does a call to x86 interrupt routine specified by <i>intnr</i>. Before call
is done it loads processor registers AX, BX, CX, DX, DI, SI and FLAGS with values from
internal register data structure. After call it saves all registers into iternal structure
and restores original values of processor registers.
@param intnr interrupt number
@return value of AX register after interrupt call
@see setreg, getreg, int86x
*/
int int86(int intnr);

/** Do extended x86 interrupt call.
Function does a call to x86 interrupt routine specified by <i>intnr</i>. Before call
is done it loads processor registers AX, BX, CX, DX, DI, SI, DS, ES and FLAGS with values from
internal register data structure. After call it saves all registers into iternal structure
and restores original values of processor registers.
@param intnr interrupt number
@return value of AX register after interrupt call
@note Before call to this function values of processors segment registers should be
transferred to internal register structure using #segread function.
@par Example:
@code
   // interrupt call example - prints current directory path 
   main(;buf,bufptr)
   {
      buf = newstring(200);       // buffer for result 
      segread();                  // init variables for seg-registers
      setreg("ah",0x47);          // dosfunc nr.
      setreg("dl",0);             // get path for current drive 
      bufptr = addr(buf);         // get address of buffer contents
      setreg("ds",bufptr >> 16);  // high word of buffer address -> segment
      setreg("si",bufptr & 0xFF); // low word of buffer address -> offset
      int86x(0x21);               // call DOS-Interrupt
      print(buf,"\n");
   }
@endcode
@see setreg, getreg, int86, segread
*/
int int86x(int intnr);

/** @name Port I/O functions */
//@{

/** Read byte from I/O port.
Function reads single byte from I/O port specified by <i>portid</i>.
@param portid id (number) of I/O port to read byte from
@return byte read
*/
int inportb(int portid);

/** Write byte to I/O port.
Function writes single byte (lowest byte of <i>value</i>) to port <i>portid</i>.
@param portid id (number) of I/O port to write byte to
@param value value to write
@return null
*/
null outportb(int portid, int value);

/** Read word from I/O port.
Function reads 16 bits word from I/O port specified by <i>portid</i>.
@param portid id (number) of I/O port to read word from
@return word read
*/
int inport(int portid);

/** Write word to I/O port.
Function writes 16 bits word (low word of <i>value</i>) to port <i>portid</i>.
@param portid id (number) of I/O port to write word to
@param value value to write
@return null
*/
null outport(int portid, int value);
//@}

/** @name Graphic BIOS functions.
Functions of this section allow rudimentary graphic programming using BIOS interrupt calls.
If more functions needed, they may be implemented via #int86 or #int86x calls.
*/
//@{
/** Set screen mode.
Function clears the screen and switches display into specified <i>mode</i>. After call
cursor position is (0,0).
@param mode Screen mode to switch display into. Meaning of value depends on display adaptor.
Usually mode 7 specifies monochrome text mode (MDA).
@return display mode used by adaptor after call.
@see setpixel, line
*/
int setscrmode(int mode);

/** Set pixel in graphics mode.
Function sets a single graphics pixel to specified <i>color</i>.
@param x x-coordinate of pixel to set
@param y y-coordinate of pixel to set
@param color new color value to set for specified pixel
@return null
@note <ul>
  <li>Display must be switched into a graphics mode for use this function.</li>
  <li>Coordinates of pixel both should be non-negative values. Maximum possible
  coordinate values are determined by resolution of screen (depending on selected
  graphics mode.<br>The coordinate pair (0,0) specifies upper left corner of screen.</li>
  <li>Because maximum color depth of BIOS graphic modes is restricted to 125, only lowest byte 
  of <i>color</i> argument is significant.</li>
</ul>
@see setscrmode, line
*/
null setpixel(int x, int y, int color);

/** Draw a line.
Function draws a graphics line from point (<i>x1</i>,<i>y1</i>) to point (<i>x2</i>,<i>y2</i>).
@param x1 x-coordinate of start point
@param y1 y-coordinate of start point
@param x2 x-coordinate of end point
@param y2 y-coordinate of end point
@param color color to use for drawing the line
@return null
@note <ul>
<li>Display must be switched into a graphics mode for use this function.</li>
<li>Coordinates of pixel both should be non-negative values. Maximum possible
coordinate values are determined by resolution of screen (depending on selected
graphics mode.<br>The coordinate pair (0,0) specifies upper left corner of screen.</li>
<li>Because maximum color depth of BIOS graphic modes is restricted to 125, only lowest byte 
of <i>color</i> argument is significant.</li>
</ul>
@par Example for using graphic functions:
@code
   // fill screen with lines
   main()
   {
      // Resolution is restricted to 240x64 pixels here. This allows running the
      // sample on ATARI-Portfolio too.
      px=240;  // pixels horizontal
      py=64;   // pixels vertical
      // switch to graphic mode  
      setscrmode(4);  // should work on most display adapters
      // get current time 
      ms = timer();
      n=0;
      // draw some lines 
      for (x=0;x<px;x+=2)
      {
         n++;
         x2= px -x;
         line(x,0,x2,py-1,1);
      }
      for (y=0;y<py;y+=2)
      {
         n++;
         y2= py-1 -y;
         line(0,y,px-1,y2,2);
      }
      // get elapsed time 
      ms = timer() - ms;
      r = n*1000 / ms;
      gets();
      // switch back to text mode 
      setscrmode(7);
      print(r, " lines per second\n");
   }
@endcode
@see setscrmode, setpixel
*/
null line(int x1, int y1, int x2, int y2, int color);
//@}

/// @}

/// @}


/// @}
