
#include "bppdef.h"
#include "bppbuf.h"
#include "bppufunc.h"
#include <stdlib.h>

/// a buffer at least can contain one element
static const size_t MINCAPACITY = 1;

/**
Default constructor allocates memory for a buffer of minimum capacity. Elements are initialized to
null.
@see MINCAPACITY
*/
BppBuffer::BppBuffer()
:_size(0),_capacity(MINCAPACITY)
{
	_data = (BppByte*) allocmem(_capacity);
	memset(_data,0,_capacity);
}

/**
This constructor allocates memory for a buffer of specified capacity. Elements are initialized to
null.
@param capacity initially capacity of buffer in bytes
@see MINCAPACITY
*/
BppBuffer::BppBuffer(size_t capacity)
:_size(0),_capacity(capacity>=MINCAPACITY ? capacity : MINCAPACITY)
{
	_data = (BppByte*) allocmem(_capacity);
	if (_data != 0)
		memset(_data,0,_capacity);
}

/**
@param cpy buffer instance to create copy from
*/
BppBuffer::BppBuffer(const BppBuffer& cpy)
:_size(cpy._size),_capacity(cpy._capacity)
{
	_data = (BppByte*) allocmem(_capacity);
	if (_data)
	  memcpy(_data,cpy._data,_capacity);
}

BppBuffer::~BppBuffer()
{
	if (_data != 0)
		freemem(_data);
}

/**
@return new copy of current buffer
@note Caller is responsible for delete result if no longer needed.
*/
BppBuffer* BppBuffer::clone() const
{
	return new BppBuffer(*this);
}

/**
@param assignment buffer instance to assign data from
@return Self-reference of buffer object.
@note <ul>
<li>Operator makes a copy of all data stored in <i>assignmant</i>.</li>
<li>Address of internal data buffer may be changed by this operator.</li>
</ul>
*/
BppBuffer& BppBuffer::operator=(const BppBuffer& assignment)
{
	if (&assignment == this)
		return *this;
	BppByte* newdata = (BppByte*) allocmem(assignment._capacity);
	if (newdata) memcpy(newdata,assignment._data,assignment._capacity);
	freemem(_data);
	_data = newdata;
	_capacity = assignment._capacity;
	_size = assignment._size;
	return *this;
}



/**
@param pos Index of word in buffer.
@return Word (16 bits) at position <i>pos</i>.
*/
BppWord BppBuffer::getW(size_t pos) const
{
#ifndef WINCE
	return *(BppWord*)(&(_data[pos]));
#else
	BppWord result;
	memcpy(&result,&(_data[pos]),sizeof(BppWord));
	return result;
#endif
}

/**
@param pos Index of dword in buffer.
@return DWord (32 bits) at position <i>pos</i>.
*/
BppDWord BppBuffer::getDW(size_t pos) const
{
#ifndef WINCE
	return *(BppDWord*)(&(_data[pos]));
#else
	BppDWord result;
	memcpy(&result,&(_data[pos]),sizeof(BppDWord));
	return result;
#endif
}

/**
@param pos Index of native word in buffer.
@return Native word (16 or 32 bits, depending on platform) at position <i>pos</i>.
*/
BppNWord BppBuffer::getNW(size_t pos) const
{
#ifndef WINCE
	return *(BppNWord*)(&(_data[pos]));
#else
	BppNWord result;
	memcpy(&result,&(_data[pos]),sizeof(BppNWord));
	return result;
#endif
}

/**
@param pos Index of word in buffer.
@param val Value to write at position <i>pos</i> into buffer.
*/
void BppBuffer::setW(size_t pos, BppWord val)
{
	memcpy(&(_data[pos]),&val,sizeof(val));
}

/**
@param pos Index of dword in buffer.
@param val Value to write at position <i>pos</i> into buffer.
*/
void BppBuffer::setDW(size_t pos, BppDWord val)
{
	memcpy(&(_data[pos]),&val,sizeof(val));
}

/**
@param pos Index of native word in buffer.
@param val Value to write at position <i>pos</i> into buffer.
*/
void BppBuffer::setNW(size_t pos, BppNWord val)
{
	memcpy(&(_data[pos]),&val,sizeof(val));
}


/**
Function changes capacity of buffer to specified number of elements. If new capacity is larger than
old one, additional allocated mamory is initialized with NULL bytes.
@param requested new capacity
@return really available new capacity after resizing
@note Function never discards elements managed by buffer, i.e. if <i>capacity</i> is lower than
current size (result of size() method) new capacity will be equal to size.
*/
size_t BppBuffer::setCapacity(size_t capacity)
{
	if (capacity == 0) capacity = MINCAPACITY;
	if (capacity < _size)
	{
		capacity = _size;
	}
	void* p = reallocmem(_data, capacity);
	_data = (BppByte*)p;
	if (capacity > _capacity)
	{
		memset(&_data[_capacity], 0, capacity - _capacity);
	}
	_capacity = capacity;
	return _capacity;
}


/**
Function adjusts capacity to current size, i.e. it releases unused memory.
@return new capacity
*/
size_t BppBuffer::shrink()
{
	return setCapacity(_size);
}

/**
Operator creates a new buffer object containing concatenation of own value and value
of <i>app</i>. Capacity (and size) of result is the sum of sizes of both operands.
@param app other buffer to concatenate with this one
@return new buffer containing concatenation of operands
*/
BppBuffer BppBuffer::operator+(const BppBuffer& app)
{
	BppBuffer tmp(*this);
	return tmp += app;
}

/**
Operator appends contents of <i>app</i> to current one.
@param app buffer to append contents from
@return reference to this object
@note If current capacity is lower than sum of both buffer sizes, it is enlarged to this sum.
*/
BppBuffer& BppBuffer::operator+=(const BppBuffer& app)
{
	size_t newsize = _size + app._size;
	if (_capacity < newsize)
		setCapacity(newsize);
	memcpy(_data+_size,app._data,app._size);
	_size += app._size;
	return *this;
}

/**
Function removes last byte from buffer.
@return byte removed.
@note This function does not affect capacity. Memory space of removed byte is set to null.
*/
BppByte BppBuffer::pop_back()
{
  if (_size == 0)
	  return 0;
  BppByte tmp = _data[--_size];
  _data[_size] = 0;
  return tmp;
}

/**
Function removes last word (i.e. two bytes) from buffer.
@return word removed.
@note This function does not affect capacity. Memory space of removed word is set to null.
*/
BppWord BppBuffer::pop_backW()
{
	BppWord w = pop_back() << 8;
	w |= pop_back();
	return w;
}

/**
Function removes last double word (i.e. four bytes) from buffer.
@return double word removed.
@note This function does not affect capacity. Memory space of removed double word is set to null.
*/
BppDWord BppBuffer::pop_backDW()
{
	BppDWord dw = pop_backW() << 16;
	dw |= pop_backW();
	return dw;
}

/**
Function removes last native word from buffer. Number of bytes removed depends on word size
of machine (or operating system). On a 16-Bit machine this function is same as pop_backW where
on a 32-Bit machine same as pop_backDW.
@return native word removed.
@note This function does not affect capacity. Memory space of removed native word is set to null.
*/
BppNWord BppBuffer::pop_backNW()
{
#if (!defined(GCC) && (INT_MAX == SHRT_MAX))
	return pop_backW();
#else
	return pop_backDW();
#endif
}

/**
Function pushes (i.e. appends) a single byte to end of buffer.
@param b byte to push
@return size of buffer before appending <i>b</i>
*/
size_t BppBuffer::push_back(BppByte b)
{
	if (_size == _capacity)
		setCapacity(2*_size);
	_data[_size] = b;
	return (size_t) _size++;
}

/**
Function pushes (i.e. appends) a word (two bytes) to end of buffer.
@param w word to push
@return size of buffer (in bytes) before appending <i>w</i>
*/
size_t BppBuffer::push_backW(BppWord w)
{
	push_back((BppByte)w);
	push_back(w >> 8);
	return (size_t) (_size -2);
}

/**
Function pushes (i.e. appends) a double word (four bytes) to end of buffer.
@param dw double word to push
@return size of buffer (in bytes) before appending <i>dw</i>
*/
size_t BppBuffer::push_backDW(BppDWord dw)
{
	push_backW((BppWord)dw);
	push_backW(dw >> 16);
	return (size_t) (_size-4);
}

/**
Function pushes (i.e. appends) a native word to end of buffer. Number of bytes appended depends on
word size of machine (or operating system). On a 16-Bit machine this function is same as push_backW where
on a 32-Bit machine same as push_backDW.
@param nw native word to push
@return size of buffer (in bytes) before appending <i>nw</i>
*/
size_t BppBuffer::push_backNW(BppNWord nw)
{
#if (!defined(GCC) && (INT_MAX == SHRT_MAX))
	return push_backW(nw);
#else
	return push_backDW(nw);
#endif
}

/**
Function removes all managed elements of buffer, i.e. it sets current size to 0 and overwrites
all elements with null value.
*/
void BppBuffer::clear()
{
	if ((_data != 0) && (_size != 0))
		memset(_data,0,_size);
	_size = 0;
}



/**
Function simply sets internal _size variable without changing any data nor capacity.
@param size new size of buffer
@return true on sucess, false if <i>size</i> is greater than current capacity
*/
bool BppBuffer::setSize(size_t size)
{
	if (size > _capacity) return false;
	_size = size;
	return true;
}


int BppBuffer::compare(const BppBuffer& comparison) const BPPNOEXCEPT
{
	if (&comparison == this) return true;
	if (_size == 0) return comparison._size ? -1 : 0;
	if (comparison._size == 0) return 1;
	size_t s = _size > comparison._size ? comparison._size : _size;
	int r = memcmp(_data,comparison._data,s);
	if (r != 0) return r;
	return _size == comparison._size ? 0 : s < _size ? 1 : -1; 
}

bool BppBuffer::operator==(const BppBuffer& cmp) const BPPNOEXCEPT
{
	if (&cmp == this) return true;
	if (_size != cmp._size) return false;
	if (_size == 0) return true;
	 return memcmp(_data,cmp._data,_size) == 0;
}

bool BppBuffer::operator!=(const BppBuffer& cmp) const BPPNOEXCEPT
{
	if (&cmp == this) return false;
	if (_size != cmp._size) return true;
	if (_size == 0) return false;
	 return memcmp(_data,cmp._data,_size) != 0;
}

/**
Default constructor allocates memory for a buffer of minimum capacity. Elements are initialized to
null character.
@see MINCAPACITY
*/
BppCharBuffer::BppCharBuffer()
:_size(0),_capacity(MINCAPACITY)
{
	_data = (BppChar*) allocmem(_capacity * sizeof(BppChar));
	memset(_data,0,_capacity * sizeof(BppChar));
}

/**
This constructor allocates memory for a buffer of specified capacity. Elements are initialized to
null character.
@param capacity initially capacity of buffer in characters
@see MINCAPACITY
*/
BppCharBuffer::BppCharBuffer(size_t capacity)
:_size(0),_capacity(capacity>=MINCAPACITY ? capacity : MINCAPACITY)
{
	_data = (BppChar*) allocmem(_capacity * sizeof(BppChar));
	memset(_data,0,_capacity * sizeof(BppChar));
}

/**
@param cpy buffer instance to create copy from
*/
BppCharBuffer::BppCharBuffer(const BppCharBuffer& cpy)
:_size(cpy._size),_capacity(cpy._capacity)
{
	_data = (BppChar*) allocmem(_capacity * sizeof(BppChar));
	if (_data) memcpy(_data,cpy._data,_capacity * sizeof(BppChar));
	//memset(_data,0,_capacity * sizeof(BppChar));
	//for (size_t i=0;i<_size;i++)
	//   _data[i] = cpy[i];
}

BppCharBuffer::BppCharBuffer(const BppChar* cstr)
:_size(0),_capacity(MINCAPACITY)
{
	size_t len = cstr ? _tcslen(cstr) : 0;
	if (len > 0) 
	{
		_capacity = len+1;
		_size = len;
		_data = (BppChar*) allocmem(_capacity * sizeof(BppChar));
		_tcscpy(_data,cstr);
	}
	else
	{
		_data = (BppChar*) allocmem(_capacity * sizeof(BppChar));
		memset(_data,0,_capacity * sizeof(BppChar));
	}
}


BppCharBuffer::~BppCharBuffer()
{
	if (_data != 0)
		freemem(_data);
}

/**
@return new copy of current buffer
@note Caller is responsible for delete result if no longer needed.
*/
BppCharBuffer* BppCharBuffer::clone() const
{
	return new BppCharBuffer(*this);
}
/**
@param assignment character buffer instance to assign data from
@return Self-reference of buffer object.
@note <ul>
<li>Operator makes a copy of all data stored in <i>assignmant</i>.</li>
<li>Address of internal data buffer may be changed by this operator.</li>
</ul>
*/
BppCharBuffer& BppCharBuffer::operator=(const BppCharBuffer& assignment)
{
	if (&assignment == this)
		return *this;
	BppChar* newdata = (BppChar*) allocmem(assignment._capacity * sizeof(BppChar));
	if (newdata) memcpy(newdata,assignment._data,assignment._capacity * sizeof(BppChar));
	freemem(_data);
	_data = newdata;
	_capacity = assignment._capacity;
	_size = assignment._size;
	return *this;
}

/**
@param cstr C character string to assign data from
@return Self-reference of buffer object.
@note <ul>
<li>Operator makes a copy of all characters stored in <i>cstr</i>.</li>
<li>Address of internal data buffer may be changed by this operator.</li>
</ul>
*/
BppCharBuffer& BppCharBuffer::operator=(const BppChar* cstr)
{
	if (cstr == _data)
		return *this;
	size_t len = cstr ? _tcslen(cstr) : 0;
	if (!len) 
	{
		 clear();
		return *this;
	}
	setCapacity(len+1);
	_tcscpy(_data,cstr);
	_size = len;
	return *this;
}

/**
@param index Index of character to access to. The index must be lower than current capacity.
@return writable reference to specified character
@note Operator directly accesses the size_ternal data buffer without any range checking.
*/
BppChar& BppCharBuffer::operator[](size_t index)
{
	return _data[index];
}

/**
@param index Index of character to access to. The index must be lower than current capacity.
@return value of specified character
@note Operator directly accesses the size_ternal data buffer without any range checking.
*/
BppChar BppCharBuffer::operator[](size_t index) const
{
	return _data[index];
}

/**
@return number of characters currently contained in buffer
*/
size_t BppCharBuffer::size() const BPPNOEXCEPT
{
	return _size;
}

/**
@return maximum number of characters buffer can hold without resizing
*/
size_t BppCharBuffer::capacity() const BPPNOEXCEPT
{
	return _capacity;
}

/**
Function changes capacity of buffer to specified number of characters. If new capacity is larger than
old one, additional allocated memory is initialized with NULL characters.
@param requested new capacity
@return really available new capacity after resizing
@note Function never discards any characters contained in buffer, i.e. if <i>capacity</i> is lower than
current size (result of size() method) new capacity will be equal to size.
*/
size_t BppCharBuffer::setCapacity(size_t capacity)
{
	if (capacity == 0) capacity = MINCAPACITY;
	if (capacity == _capacity) return _capacity;
	if (capacity < _size)
	{
		capacity = _size;
	}
	void* p = reallocmem(_data, capacity * sizeof(BppChar));
	_data = (BppChar*)p;
	if (capacity > _capacity)
	{
		memset(&_data[_capacity], 0, (capacity - _capacity) * sizeof(BppChar));
	}
	_capacity = capacity;
	return _capacity;
}


/**
Function adjusts capacity to current size, i.e. it releases unused memory.
@return new capacity
*/
size_t BppCharBuffer::shrink()
{
	return setCapacity(_size);
}


/**
Operator creates a new character buffer containing concatenation of own value and value
of <i>app</i>. Capacity (and size) of result is the sum of sizes of both operands.
@param app other character buffer to concatenate with this one
@return new character buffer containing concatenation of operands
*/
BppCharBuffer BppCharBuffer::operator+(const BppCharBuffer& app)
{
	BppCharBuffer tmp(*this);
	return tmp += app;;
}

/**
Operator appends contents of <i>app</i> to current character buffer.
@param app character buffer to append contents from
@return reference to this object
@note If current capacity is lower than sum of both buffer sizes, it is enlarged to this sum.
*/
BppCharBuffer& BppCharBuffer::operator+=(const BppCharBuffer& app)
{
	size_t newsize = _size + app._size;
	if (_capacity < newsize)
		setCapacity(newsize);
	memcpy(_data+_size*sizeof(BppChar),app._data,app._size * sizeof(BppChar));
	//for (size_t i=0;i<app._size;i++)
	//   _data[_size+i] = app._data[i];
	_size += app._size;
	return *this;
}

/**
Function removes last character from buffer.
@return character removed.
@note This function does not affect capacity. Memory space of removed character is set to null.
*/
BppChar BppCharBuffer::pop_back()
{
  if (_size == 0)
	  return 0;
  _size--;
  BppChar tmp = _data[_size];
  _data[_size] = 0;
  return tmp;
}

/**
Function pushes (i.e. appends) a single character to end of buffer.
@param b character to push back
*/
void BppCharBuffer::push_back(BppChar b)
{
	if (_size >= _capacity)
		setCapacity(2*_size);
	_data[_size] = b;
	++_size;
}


/**
Function removes all characters from buffer, i.e. it sets current size to 0 and overwrites
all characters with null value.
*/
void BppCharBuffer::clear()
{
	if ((_data != 0) && (_size != 0))
		memset(_data,0,_size*sizeof(BppChar));
	_size = 0;
}

/**
@return posize_ter to size_ternal data area
@note This function is mainly for size_ternal use. Client code must not delete nor resize returned data space.<br>
Implementation of class ensures returned pointer to be usable like a C-style character string.
*/
BppChar* BppCharBuffer::data() const BPPNOEXCEPT
{
	return _data;
}

/**
Function sets internal size attribute to number of characters of C-string at start of internal buffer.
It is intended for internal use only.
*/
void BppCharBuffer::adjustSize()
{
	if (_data != 0)
	{
		_data[_capacity-1] = 0;
	  _size = (size_t) _tcslen(_data);
	}
	else
		_size = 0;
}

int BppCharBuffer::compare(const BppCharBuffer& comparison) const BPPNOEXCEPT
{
	if (&comparison == this) return 0;
	if (_size == 0) return comparison._size ? -1 : 0;
	if (comparison._size == 0) return 1;
	size_t s = _size > comparison._size ? comparison._size : _size;
	s *= sizeof(BppChar);
	int r = memcmp(_data,comparison._data,s);
	if (r != 0) return r;
	return _size == comparison._size ? 0 : s < _size ? 1 : -1; 
}

bool BppCharBuffer::operator==(const BppCharBuffer& cmp) const BPPNOEXCEPT
{
	if (&cmp == this) return true;
	if (_size != cmp._size) return false;
	if (_size == 0) return true;
	 return memcmp(_data,cmp._data,_size*sizeof(BppChar)) == 0;
}

bool BppCharBuffer::operator!=(const BppCharBuffer& cmp) const BPPNOEXCEPT
{
	if (&cmp == this) return false;
	if (_size != cmp._size) return true;
	if (_size == 0) return false;
	 return memcmp(_data,cmp._data,_size*sizeof(BppChar)) != 0;
}

#ifndef DOXY
BppDWord BytePtr2DW(BppByte* p)
{
#ifndef WINCE
	return *(BppDWord*) p;
#else
	BppDWord tmp;
	memcpy(&tmp,p,sizeof(tmp));
	return tmp;
#endif
}

BppWord BytePtr2W(BppByte* p)
{
#ifndef WINCE
	return *(BppWord*) p;
#else
	BppWord tmp;
	memcpy(&tmp,p,sizeof(tmp));
	return tmp;
#endif
}

BppNWord BytePtr2NW(BppByte* p)
{
#ifndef WINCE
	return *(BppNWord*) p;
#else
	BppNWord tmp;
	memcpy(&tmp,p,sizeof(tmp));
	return tmp;
#endif
}

#endif // !DOXY
