#include "rational.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
// minimum possible value of raint type (MSB 1, all other bits 0)
static const raint RINTMIN = (((raint)1) << ((sizeof(raint)*8)-1));
// maximum possible value of raint type (MSB 0, all other bits 1)
static const raint RINTMAX = (((raint)-1) ^ RINTMIN);
// half of maximum value for raint type (e.g. 0x0FFFF in case of raint is long)
static const raint HALF = (raint)(((rauint)-1) >> (sizeof(rauint)*4));

/**
@param p nominator
@param q denominator
*/ 
rational::rational(raint p, rauint q)
:_nom(p),_denom(q),_buf(0)
{
	norm();
}

/**
@param d double value to create rational from
*/
rational::rational(double d)
:_nom(d),_denom(1),_buf(0){
	double dd = (d < 0) ? -d : d;
	// make denominator between 0 and HALF
	raint scale = (double)HALF / dd; 
    if (scale <= 0) return;
	if (scale > HALF) scale = HALF;
	_nom = d * scale;
	_denom = scale;
	norm();
}

rational::rational(const rational& cpy)
:_nom(cpy._nom),_denom(cpy._denom),_buf(0){
}

/**
Ctor parses string in <i>str</i> and builds rational value.
@param str String to parse. Format of <i>str</i> may be:
<ul>
<li>integer value in decimal string representation,</li>
<li>string representation of rational (nominator/denominator) or</li>
<li>valid string representation of double value.</li>
</ul>
*/
rational::rational(const char* str)
:_nom(0),_denom(1),_buf(0)
{
	if (!str) return;
	const char* p = strchr(str,'/');
    if (p)
	{
		char* b = strdup(str);
		int i = p-str;
		b[i++] = 0;
        _nom = atoi(b);
		int den = atoi(b+i);
		if (den < 0)
		{
			_denom = -den;
			_nom = -_nom;
		}
		else
			_denom = den;
		free(b);
	    norm();
	}
	else if (p = strchr(str,'.'))
	{
		if (strchr(p,'E') || strchr(p,'e')) {
			rational tmp = atof(str);
			_nom = tmp._nom;
			_denom = tmp._denom;
		}
		else
		{
			char* b = strdup(str);
			int i = p-str;
			b[i++] = 0;
			_nom = atoi(b);
			raint sig = _nom < 0 ? -1 : 1;
			_nom *= sig;
			p = b+i;
 //           raint den = 10;
			while (char c = *p++)
			{
				if (c < '0' || c > '9') break;
				c -= '0';
				if ((_nom > HALF) || (_denom > HALF)) 
				{
					if (c >= 5) ++_nom;
					break;
				}
				_nom *= 10;
				_denom *= 10;
				if (c > 0) _nom += (raint)c;
//				den *= 10;
			}
			_nom *= sig;
			norm();
			free(b);
		}
	}
	else _nom = atoi(str);
}

/**
Destructor frees internal character buffer (if it exists).
*/
rational::~rational(){
	if (_buf) free(_buf);
}

/**
Assigns nominator ant denominator (but not internal character buffer).
@param cpy source of assignment
@return reference to own object
*/
rational& rational::operator=(const rational& cpy)
{
	if (this != &cpy)
	{
		_nom = cpy._nom;
		_denom = cpy._denom;
	}
	return *this;
}

rational::operator raint() const {
	if (!_denom) return _nom < 0 ? RINTMIN : RINTMAX;
	return _nom / _denom;
}

rational::operator double() const
{
	if (!_denom) return _nom < 0 ? -DBL_MAX  : DBL_MAX;
	return ((double)_nom) / ((double)_denom);
}

#if SHRT_MAX == INT_MAX
"%d/%d";
#else
#define fmtnum "%I64d"
#endif
static const char* fmt = fmtnum "/" fmtnum;

rational::operator const char*() const{
	if (!_denom) return "NaN";
	if (!_buf)
	{
		(char*) _buf = (char*) malloc(100);
	}
	if (_denom == 1) sprintf((char*)_buf,fmtnum,_nom);
	else sprintf((char*)_buf,fmt,_nom,_denom);
	return _buf;
}

const char* rational::fltStr(size_t maxdec) const{
	if (!_denom) return "NaN";
	if (!_buf)
	{
		(char*) _buf = (char*) malloc(100);
	}
	char* p = (char*) _buf;
	raint n = _nom;
    raint d = _nom / (raint)_denom;
	if (n < 0) 
	{
		*(p++) = '-';
		d = -d;
		n = -n;
	}
    sprintf(p,fmtnum,d);
	rauint rem = n % _denom;
	if (rem > 0 && maxdec > 0)
	{
		size_t l = strlen(_buf);
		if (maxdec > (98-l)) maxdec = (98-l);
		p = (char*)_buf + l;
		*(p++) = '.';
		do {
			d = (10*rem) / _denom;
			rem = (10*rem) % _denom;
			*(p++) = '0'+d;
		} while ((rem > 0) && (--maxdec > 0));
        *p = 0;
	}
	return _buf;
}

rational rational::reduce(size_t maxdec) const
{
	return rational(fltStr(maxdec));
	//rational res = *this;
	//int sig = 1;
	//if (_nom < 0) { sig = -1; res._nom = -_nom;}
	//while (res._nom > HALF && res._denom > HALF ) 
	//  { res._nom >>= 1; res._denom >>= 1;}
	//if (sig < 0) res._nom = -res._nom;
	//return res;
}

rational& rational::operator+=(const rational& r){
	if (_denom == r._denom) _nom += r._nom;
	else {
		rauint rdenom = lcm(_denom,r._denom);
		if (rdenom)
		{
			_nom = _nom * (rdenom / _denom) + 
				r._nom * (rdenom / r._denom);
			_denom = rdenom;
		}
		else if (_nom == 0) (*this) = r; 
	}
	norm();
	return *this;
}

rational& rational::operator-=(const rational& r){
	if (_denom == r._denom) _nom -= r._nom;
	else {
		rauint rdenom = lcm(_denom,r._denom);
		if (rdenom) {
			_nom = _nom * (rdenom / _denom) - 
				r._nom * (rdenom / r._denom);
			_denom = rdenom;
		 }
		else if (_nom == 0) {_nom = -r._nom; _denom = r._denom; } 
	}
	norm();
	return *this;
}

rational& rational::operator*=(const rational& r){
	_nom *= r._nom;
	_denom *= r._denom;
	norm();
	return *this;
}

rational& rational::operator/=(const rational& r){
	_nom *= r._denom;
	_denom *= r._nom;
	norm();
	return *this;
}

raint rational::compare(const rational& cmp) const{
	if (_nom==cmp._nom && _denom==cmp._denom) return 0;
	return ((*this) - cmp)._nom;
}


rauint rational::gcd(rauint a, rauint b){
	while (b) { raint c = a % b; a = b; b = c; }
	return a > 0 ? a : 1;
}

rauint rational::lcm(rauint a, rauint b){
	return a*b / gcd(a,b);
}


void rational::norm(){
	rauint q = gcd(_nom < 0 ? -_nom : _nom,_denom);
	if (q > 1)
	{
		_nom /= (raint) q;
        _denom /= q;
	}
}

rational rational::pow_n(unsigned n) const
{
	if (n == 0) return (raint)1;
	return pow((double)*this,(double)n);
	//rational result = *this;
	//rational x = *this;
	//if (riabs(x._nom) > HALF || x._denom > HALF)
	//	x = x.reduce();
	//for (unsigned i=1;i<n;++i) 
	//{
	//	if (riabs(result._nom) > HALF || result._denom > HALF)
	//		result = result.reduce();
	//	result *= x;
	//}
	//if (riabs(result._nom) > HALF || result._denom > HALF)
	//		return result.reduce();
	//return result;
}

rational rational::sqrt() const
{
	return ::sqrt((double)*this);
	//static const rational half(1,2);
	//static const rational eps(1,HALF);
	//static const int maxloop = 100;
	//rational r(1,1);
	//int loop = maxloop;
	//do {
	//   r = half * (r + ((*this)/r));
	//   //if (r._nom > HALF || (r._denom > HALF)) break;
	//} while (--loop && (((r*r) - (*this)).abs() > eps));
 //   return r;
}


rational rational::root(unsigned n) const
{
	// x = rootn(a,n) : x --> ((n-1)*x + a/pow(x,n-1))/n
	//static const int maxloop = 100;
	//static const rational eps(1,HALF);
	if (n == 0) return raint(1);
	if (n == 1) return *this;
	return pow((double)*this,1.0/n);
//	rational rn = (raint)n;
//	rational rn1 = (raint)(n-1);
//	rational r = raint(1);
//	int l = 0;
//	do {
//		if (++l > maxloop) break;
//	   if (riabs(r._nom) > HALF || (r._denom > HALF)) break;
//		r = (rn1*r + (*this)/r.pow_n(n-1)) / rn; 
////		rational denom1 = r.pow_n(n-1) * rn; 
//	  // r -= (r.pow_n(n) - *this) / denom1;
//	} while (((r.pow_n(n)) - (*this)).abs() > eps);
//    return r;
}

rational rational::abs() const
{
	return rational(_nom < 0 ? -_nom : _nom,_denom);
}

rational operator+(const rational& l, const rational& r){
	return (rational(l)+=r);
}

rational operator-(const rational& l, const rational& r){
	return (rational(l)-=r);
}

rational operator*(const rational& l, const rational& r){
	return (rational(l)*=r);
}

rational operator/(const rational& l, const rational& r){
	return (rational(l)/=r);
}

bool operator==(const rational& l, const rational& r)
{
	return l.compare(r) == 0;
}
bool operator!=(const rational& l, const rational& r)
{
	return l.compare(r) != 0;
}
bool operator<=(const rational& l, const rational& r)
{
	return l.compare(r) <= 0;
}
bool operator>=(const rational& l, const rational& r)
{
	return l.compare(r) >= 0;
}
bool operator<(const rational& l, const rational& r)
{
	return l.compare(r) < 0;
}
bool operator>(const rational& l, const rational& r)
{
	return l.compare(r) > 0;
}
