#ifndef RATIONAL_H
#define RATIONAL_H
#include <limits.h>

#if SHRT_MAX == INT_MAX
typedef long raint;
typedef unsigned long rauint;
#else
typedef long long raint;
typedef unsigned long long rauint;
#endif


inline raint riabs(raint v) { return v<0 ? v : -v;}
class rational
{
public:
	/// Initialization from integer type value(s).
	rational(raint p, rauint q=1);
	///Initialization from double value.
	rational(double d);
	/// Initialization from character string.
    rational(const char* str);
	/// Copy ctor
    rational(const rational& cpy);
	// dtor
	~rational();
	/// assignment
	rational& operator=(const rational& cpy);
    /// Conversion to integer type
	operator raint() const ;
	/// Conversion to double
	operator double() const;
	/// Conversion to character string.
	operator const char*() const;
	/// Conversion to string in floating point format.
    const char* fltStr(size_t maxdec = sizeof(raint)) const; 
	/// addition/assignment
    rational& operator+=(const rational& r);
	/// subtraction/assignment
	rational& operator-=(const rational& r);
	/// multiplication/assignment
	rational& operator*=(const rational& r);
	/// division/assignment
	rational& operator/=(const rational& r);
	/// comparison method
    raint compare(const rational& cmp) const;
	/// Computes own value power to n.
    rational pow_n(unsigned n) const;
	/// Computes square root.
    rational sqrt() const;
	/// Computes n-th root.
    rational root(unsigned n) const;
	/// Gets absolute value.
	rational abs() const;
	/// Reduced internal representation.
	rational reduce(size_t maxdec = sizeof(raint)) const;
 private:
	 /// greatest common divisor of two integers
    static rauint gcd(rauint a, rauint b);
	 /// lowest common multiple of two integers
    static rauint lcm(rauint a, rauint b);

	/// Normalizes internal representation.
	void norm();
	/// nominator
	raint _nom;
	/// denominator
	rauint _denom;
	/// Buffer for building string representations.
	char* _buf;
};

/// addition
rational operator+(const rational& l, const rational& r);
/// subtraction
rational operator-(const rational& l, const rational& r);
/// multiplication
rational operator*(const rational& l, const rational& r);
/// division
rational operator/(const rational& l, const rational& r);
/// equal
bool operator==(const rational& l, const rational& r);
/// not equal
bool operator!=(const rational& l, const rational& r);
/// lower or equal
bool operator<=(const rational& l, const rational& r);
/// greater or equal
bool operator>=(const rational& l, const rational& r);
/// lower than
bool operator<(const rational& l, const rational& r);
/// greater than
bool operator>(const rational& l, const rational& r);

#endif // RATIONAL.H