/*
 * Copyright (c) 2001-2002 Daniel van Vugt <danv@users.sf.net>
 * This file is part of the Rex library.
 * See LICENSE.txt for license details.
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "rex.h"

#ifdef _WIN32
#define DIRSLASH '\\'
#else
#define DIRSLASH '/'
#endif

typedef enum {NOOP, FIND, MATCH, CHECK} op_t;

static const char title[] = "RAT (Rex Auto-Tester) v0.03";


static void
ReportError(const char *filename, int line, int offset, const char *msg)
{
	fprintf(stderr, "Error: %s at %s line %d pos %d\n", msg, filename, line, offset+1);
}


static int
Keyword(const char *text, const char *templ)
{
	int len, result;

	len = strlen(templ);
	result = 0;
	if ( !strncmp(text, templ, len) )
	{
		char next;

		next = *(text + len);
		if ( isspace(next) || (next == '\0') )
		{
			result = len;
		}
	}

	return result;
}


static int
Execute(rex_t rex,
	rex_error_t error,
	size_t erroroffset,
	const char *expr,
	size_t exprlen,
	int sign,
	op_t op,
	const char *text,
	size_t textlen,
	int haveoffset,
	size_t atoffset,
	int havelen,
	size_t atlen,
	const char *filename,
	int linenum,
	int segments
	)
{
	const char *desc;
	int result;
	size_t answer, where;
	unsigned int options;
	static const char * const errorstring[] = REX_ERROR_STRINGS;

	switch (op)
	{
		case MATCH:
			desc = "MATCH";
			break;
		case FIND:
			desc = "IN";
			break;
		case CHECK:
			desc = "VALID";
			break;
		default:
			desc = "???";
			break;
	}

	if (segments == 1)
	{
		printf("|------------------\n"
			"| Case:     %s:%d\n"
			"| Expr:     `%s' %s%s",
			filename, linenum,
			expr, sign?"":"NOT ", desc);

		if (op != CHECK)
		{
			printf(" `%s'", text);
		}
	
		if (haveoffset)
		{
			printf(" AT %lu", (unsigned long)atoffset);
		}
	
		if ( (op == FIND) && havelen )
		{
			printf(" length %lu", (unsigned long)atlen);
		}
		printf("\n");
	}
	printf("| Segments: %d\n| Executed: ", segments);

	switch (op)
	{
		case MATCH:
			if (rex != REX_INVALID)
			{
				rex_search_t search;
				int i;
				const char *subtext;
				size_t sublen;

				options = REX_MATCH_WHOLE;
				search = rexPreallocSearch(rex);
				i = 0;
				subtext = text;
				sublen = textlen / segments;
				answer = 0;
				while (i < segments)
				{
					if (i == (segments-1))
					{
						sublen += (text + textlen) - (subtext + sublen);
						options |= REX_MATCH_EOL;
					}
					answer = rexMatch(rex, subtext, sublen, options, search, &where);
					subtext += sublen;
					i++;
				}
				rexFreeSearch(search);
				printf("rexMatch returned %lu (%s", (unsigned long)answer, answer?"matched":"no match");
				result = answer;
				if (!answer)
				{
					printf(" starting at %lu", (unsigned long)where);
				}
				printf(")");
				result = sign ? result : !result;
				if (haveoffset)
				{
					result = result && (where == atoffset);
				}
			}
			else
			{
				printf("wasn't called because ");
				result = 0;
			}
			break;

		case FIND:
			if (rex != REX_INVALID)
			{
				rex_search_t search;
				int i;
				const char *subtext;
				size_t sublen;

				options = REX_MATCH_SUBSTRING;
				options |= (!haveoffset && !havelen) ? REX_MATCH_ANY : 0;
				search = rexPreallocSearch(rex);
				i = 0;
				answer = 0;
				where = 0;
				subtext = text;
				sublen = textlen / segments;
				while (i < segments)
				{
					if (i == (segments-1))
					{
						sublen += (text + textlen) - (subtext + sublen);
						options |= REX_MATCH_EOL;
					}
					answer = rexMatch(rex, subtext, sublen, options, search, &where);
					subtext += sublen;
					i++;
				}
				rexFreeSearch(search);
				printf("rexMatch returned %lu (%sfound", (unsigned long)answer, answer?"":"not ");
				if (answer)
				{
					printf(" at offset %lu length %lu",
						(unsigned long)where, (unsigned long)answer-1);
				}
				printf(")");
				result = answer;
				result = sign ? result : !result;
				if (haveoffset)
				{
					result = result && (where == atoffset);
				}
				if (havelen)
				{
					result = result && ((answer-1) == atlen);
				}
			}
			else
			{
				printf("wasn't called because ");
				result = 0;
			}
			break;

		case CHECK:
			result = (rex != REX_INVALID);
			result = sign ? result : !result;
			if ( haveoffset && !sign )
			{
				result = result && (erroroffset == atoffset);
			}
			break;

		default:
			desc = "???";
			result = 0;
			break;
	}

	if (rex == REX_INVALID)
	{
		printf("rexCompile failed with error %d at offset %lu (%s)",
				error, (unsigned long)erroroffset, errorstring[error]);
	}
	else if (op == CHECK)
	{
		printf("rexCompile succeeded");
	}
	
	printf("\n%c Result:   %s\n", result?'|':'#', result?"PASS":"FAIL");

	return result ? 1 : 0;
}


static int
ExecuteAll(rex_t rex,
	rex_error_t error,
	size_t erroroffset,
	const char *expr,
	size_t exprlen,
	int sign,
	op_t op,
	const char *text,
	size_t textlen,
	int haveoffset,
	size_t atoffset,
	int havelen,
	size_t atlen,
	const char *filename,
	int linenum
	)
{
	int nseg, maxnseg, allok;

	allok = 1;
	maxnseg = (op == CHECK) ? 1 : textlen;
	for (nseg=1; nseg <= maxnseg; nseg++)
	{
		allok &= Execute(rex, error, erroroffset, expr, exprlen, sign,
				op, text, textlen, haveoffset, atoffset,
				havelen, atlen, filename, linenum, nseg);
	}

	return allok;
}


						
static int
Test(FILE *file, const char *name, int *total)
{
	typedef enum {CLASS, FILENAME, EXPR, SEARCH, NEWSEARCH_OR_CLASS, TEXT, CONCLUSION} expect_t;
	expect_t expect;
	op_t op;
	char buf[256];
	char expr[256];
	char text[256];
	char *p;
	int lines;
	int sign;
	int haveoffset;
	int havelen;
	size_t exprlen;
	size_t textlen;
	unsigned long atoffset, atlen;
	int passes;
	rex_t rex;
	rex_error_t error;
	size_t erroroffset;
	int count;
	int srcline;
	unsigned int compoptions;

	rex = REX_INVALID;
	expect = CLASS;
	lines = 0;
	passes = 0;
	count = 0;
	srcline = 0;
	op = NOOP;
	sign = 1;
	haveoffset = 0;
	havelen = 0;
	exprlen = 0;
	textlen = 0;
	compoptions = 0;

	while ( fgets(buf, sizeof(buf), file) != NULL )
	{
		lines++;
		p = buf;

		while (*p != '\0')
		{
			char sep;
			int len;
			FILE *file;
			char *slash;

			while ( isspace(*p) ) p++;
			if (*p == '\0')
			{
				break;
			}
			if (*p == '#')
			{
				break;
			}

			switch (expect)
			{
				case CLASS:
					if (rex != REX_INVALID)
					{
						rexFree(rex);
						rex = REX_INVALID;
					}
					if ( (len = Keyword(p, "dvregex")) )
					{
						p += len;
						expect = EXPR;
						compoptions = REX_COMPILE_DV;
					}
					else if ( (len = Keyword(p, "dvregexi")) )
					{
						p += len;
						expect = EXPR;
						compoptions = REX_COMPILE_DV | REX_COMPILE_ICASE;
					}
					else if ( (len = Keyword(p, "include")) )
					{
						p += len;
						expect = FILENAME;
					}
					else
					{
						ReportError(name, lines, p-buf, "Unknown keyword");
						expect = CLASS;
						goto nextline;
					}
					break;

				case FILENAME:
					slash = strrchr(name, DIRSLASH);
					if (slash != NULL)
					{
						exprlen = slash - name + 1;
						strncpy(expr, name, exprlen);
					}
					else
					{
						exprlen = 0;
					}
					sep = *p;
					do
					{
						p++;
						expr[exprlen] = *p;
						exprlen++;
						if (*p == '\0')
						{
							ReportError(name, lines, p-buf, "Unterminated filename");
							expect = CLASS;
							goto nextline;
						}
					} while (*p != sep);
					p++;
					exprlen--;
					expr[exprlen] = '\0';
					file = fopen(expr, "r");
					if (file != NULL)
					{
						int subtotal;
						
						passes += Test(file, expr, &subtotal);
						count += subtotal;
						fclose(file);
					}
					else
					{
						fprintf(stderr, "Failed to open file %s included by %s.\n", expr, name);
					}
					expect = CLASS;
					break;

				case EXPR:
					exprlen = 0;
					sep = *p;
					do
					{
						p++;
						expr[exprlen] = *p;
						exprlen++;
						if (*p == '\0')
						{
							ReportError(name, lines, p-buf, "Unterminated expression");
							expect = CLASS;
							goto nextline;
						}
					} while (*p != sep);
					exprlen--;
					expr[exprlen] = '\0';
					rex = rexCompile(expr, exprlen, compoptions, &error, &erroroffset);
					p++;
					expect = SEARCH;
					op = NOOP;
					sign = 1;
					haveoffset = 0;
					havelen = 0;
					break;

				case SEARCH:
				case NEWSEARCH_OR_CLASS:
					if ( (len = Keyword(p, "and")) )
					{
						p += len;
					}
					else if ( (len = Keyword(p, "but")) )
					{
						p += len;
					}
					else if ( (len = Keyword(p, "is")) )
					{
						p += len;
					}
					else if ( (len = Keyword(p, "not")) )
					{
						sign = 0;
						p += len;
					}
					else if ( (len = Keyword(p, "isn't")) )
					{
						sign = 0;
						p += len;
					}
					else if ( (len = Keyword(p, "doesn't")) )
					{
						sign = 0;
						p += len;
					}
					else if ( (len = Keyword(p, "matches")) )
					{
						op = MATCH;
						p += len;
						expect = TEXT;
					}
					else if ( (len = Keyword(p, "match")) )
					{
						op = MATCH;
						p += len;
						expect = TEXT;
					}
					else if ( (len = Keyword(p, "in")) )
					{
						op = FIND;
						p += len;
						expect = TEXT;
					}
					else if ( (len = Keyword(p, "invalid")) )
					{
						op = CHECK;
						sign = 0;
						p += len;
						expect = CONCLUSION;
						srcline = lines;
					}
					else if ( (len = Keyword(p, "valid")) )
					{
						op = CHECK;
						p += len;
						expect = CONCLUSION;
						srcline = lines;
					}
					else if (expect == SEARCH)
					{
						ReportError(name, lines, p-buf, "Invalid test");
						expect = CLASS;
						goto nextline;
					}
					else
					{
						expect = CLASS;
					}
					break;

				case CONCLUSION:
					if ( (len = Keyword(p, "at")) )
					{
						p += len;
						if ( sscanf(p, " %lu%n", &atoffset, &len) )
						{
							p += len;
							haveoffset = 1;
						}
						else
						{
							ReportError(name, lines, p-buf, "Invalid offset value");
							expect = CLASS;
							goto nextline;
						}
					}
					else if ( (len = Keyword(p, "length")) )
					{
						p += len;
						if ( sscanf(p, " %lu%n", &atlen, &len) )
						{
							p += len;
							havelen = 1;
						}
						else
						{
							ReportError(name, lines, p-buf, "Invalid length value");
							expect = CLASS;
							goto nextline;
						}
					}
					else
					{
						count++;
						passes += ExecuteAll(rex, error, erroroffset,
							expr, exprlen,
							sign, op, text, textlen,
							haveoffset, atoffset, havelen, atlen,
							name, srcline);
						expect = NEWSEARCH_OR_CLASS;
						havelen = 0;
						haveoffset = 0;
						sign = 1;
					}
					break;

				case TEXT:
					textlen = 0;
					sep = *p;
					do
					{
						p++;
						text[textlen] = *p;
						textlen++;
						if (*p == '\0')
						{
							ReportError(name, lines, p-buf, "Unterminated text");
							expect = CLASS;
							goto nextline;
						}
					} while (*p != sep);
					textlen--;
					text[textlen] = '\0';
					p++;
					expect = CONCLUSION;
					srcline = lines;
					break;

				default:
					assert(0);
					break;
			}
		}
		nextline: ;;
	}

	if (expect == CONCLUSION)
	{
		count++;
		passes += ExecuteAll(rex, error, erroroffset,
			expr, exprlen,
			sign, op, text, textlen,
			haveoffset, atoffset, havelen, atlen,
			name, srcline);
	}

	*total = count;

	if (rex != REX_INVALID)
	{
		rexFree(rex);
	}

	return passes;
}


int
main(int argc, char *argv[])
{
	FILE *file;
	const char *name;
	int total, passed;

	printf("%s (using rex v%u.%02u)\n",
		title, rexversion >> 8, rexversion & 0xff);
	if (argc > 1)
	{
		file = fopen(argv[1], "r");
		if (file == NULL)
		{
			printf("Couldn't find file `%s'\n"
				"Usage: %s [<testcasefile>]\n"
				"If <testcasefile> is omitted, input is taken from stdin.\n"
				"", argv[1], argv[0]);
			return 0;
		}
		name = argv[1];
	}
	else
	{
		file = stdin;
		name = "stdin";
	}

	passed = Test(file, name, &total);

	printf("======================\nPassed: %d\nFailed: %d\nOut of: %d\n",
		passed, total-passed, total);

	if (file != stdin)
	{
		fclose(file);
	}

	return total - passed;
}
