/*
 * Copyright (c) 2001-2002 Daniel van Vugt <danv@users.sf.net>
 * This file is part of the Rex library.
 * See LICENSE.txt for license details.
 */

#ifndef _REX_H
#define _REX_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stddef.h>

#define REX_MAKE_VERSION(a,b) (((a) << 8) | (b))

typedef enum
{
	REX_SUCCESS=0,
	REX_MEM,
	REX_INVALID_RANGE,
	REX_INVALID_ESCAPE,
	REX_INVALID_CHARCLASS,
	REX_INCOMPLETE_ESCAPE,
	REX_UNMATCHED_CLOSE,
	REX_POSTFIX,
	REX_BOUND_TOO_BIG,
	REX_INVALID_BOUND_CHAR,
	REX_INVALID_BOUND_RANGE,
	REX_EMPTY_BOUND,
	REX_EMPTY,
	REX_OPEN_PAREN,
	REX_OPEN_CHARSET,
	REX_OPEN_BOUND,
	REX_NERRORS
} rex_error_t;

#define REX_ERROR_STRINGS \
{ \
	"Success", \
	"Expression requires too much memory", \
	"Invalid character range", \
	"Invalid escape sequence", \
	"Invalid character class", \
	"Incomplete escape sequence", \
	"Unmatched closing parenthesis", \
	"Postfix operator does not follow an expression", \
	"Bound value too large", \
	"Invalid character in bound", \
	"Bound minimum is greater than the maximum", \
	"Bound is empty", \
	"Empty subexpression not allowed", \
	"Unclosed parentheses", \
	"Unclosed character set", \
	"Unclosed bound", \
}

typedef void *rex_t;
#define REX_INVALID NULL

typedef void *rex_search_t;
#define REX_NEW_SEARCH NULL
#define REX_INVALID_SEARCH NULL

#define REX_PATTERN_MAX 1024
#define REX_SEARCH_MAX ( (~(size_t)0) - 2 )
#define REX_DUP_MAX 255

/* rexCompile options */
#define REX_COMPILE_DV      0x0001  /* Syntax: DV Extended Regular Expressions */
#define REX_COMPILE_ICASE   0x0010  /* Modifier: ignore case */
#define REX_COMPILE_NEWLINE 0x0020  /* Modifier: don't allow dot (.) and
                                       character sets like [^...] to match a
                                       newline */

/* rexMatch options */
#define REX_MATCH_WHOLE     0x01  /* match the whole string */
#define REX_MATCH_SUBSTRING 0x02  /* match the first correct substring to the
                                     max length possible */
#define REX_MATCH_ANY       0x04  /* return ASAP; don't necessarily find the
                                     earilest or longest match */
#define REX_MATCH_EOL       0x08  /* match $ at the end of the string (implied
                                     if search==REX_NEW_SEARCH) */

#ifdef _WIN32
#ifdef REX_STATIC_LIBRARY
#define REX_API(rettype) rettype
#else
#ifdef REX_SOURCE
#define REX_API(rettype) _declspec(dllexport) rettype
#else
#define REX_API(rettype) _declspec(dllimport) rettype
#endif
#endif
#else
#define REX_API(rettype) rettype
#endif

extern REX_API(const unsigned int)  rexversion;

extern REX_API(rex_t)  rexCompile(const char *pattern, size_t len,
                               unsigned int options, rex_error_t *error,
                               size_t *erroroffset);
/*
 * rexCompile
 * Parses and compiles an expression into a more efficient internal format
 * (rex_t). Options must include exactly one of the REX_COMPILE syntax options
 * along with whatever REX_COMPILE modifier flags you want. The error and
 * erroroffset parameters are optional and may be NULL if you don't want error
 * details. Returns REX_INVALID if unsuccessful.
 */

extern REX_API(void)   rexFree(rex_t rex);
/*
 * rexFree
 * Frees an expression compiled with rexCompile.
 */

extern REX_API(size_t) rexMatch(rex_t rex, const char *input, size_t len,
                               unsigned int options, rex_search_t search,
                               size_t *offset);
/*
 * rexMatch
 * Does the actual matching, using a rex_t previously compiled with rexCompile.
 * You must specify one of the options: REX_MATCH_WHOLE or REX_MATCH_SUBSTRING.
 * No other options are mandatory. For simple searches the search parameter
 * should be REX_NEW_SEARCH. For continued searches, search is a rex_search_t
 * created with rexPreallocSearch(). If offset is non-NULL, it stores the
 * offset at which the match occured (or the earliest offset at which a match
 * did not occur). rexMatch returns 0 on no match, or the length of the
 * matched string plus one. This way, the return value can also be interpreted
 * as a boolean (match or no match). For continued (segmented) searches using
 * the search parameter, the return and offset values represent a global
 * length and offset, relatve to the first call to rexMatch().
 */

extern REX_API(rex_search_t) rexPreallocSearch(rex_t rex);
/*
 * rexPreallocSearch
 * Allocates a rex_search_t suitable for use in continuable searching (spanning
 * multiple calls to rexMatch). Returns REX_INVALID_SEARCH on error.
 */

extern REX_API(void)   rexResetSearch(rex_search_t search);
/*
 * rexResetSearch
 * Returns a rex_search_t to it's original state as returned by
 * rexPreallocSearch, suitable for starting a new continuable search without the
 * need to allocate a new rex_search_t.
 */

extern REX_API(void)   rexFreeSearch(rex_search_t search);
/*
 * rexFreeSearch
 * Frees a rex_search_t that was created with rexPreallocSearch.
 */

extern REX_API(size_t) rexLimit(size_t lim);
/*
 * rexLimit
 * Limits the size to which a rex_t can grow before rexCompile returns failure
 * with a REX_MEM error. This is necessary in order to protect against valid
 * but dangerously large expressions like:
 *   a{100}{100}{100}{100}
 * Returns the new memory limit, or if lim is zero, returns the existing limit
 * unchanged. rexLimit only affects subsequent calls to rexCompile in the same
 * process.
 */

#ifdef __cplusplus
}
#endif

#endif
