/*
 * Copyright (c) 2001-2002 Daniel van Vugt <danv@users.sf.net>
 * This file is part of the Rex library.
 * See LICENSE.txt for license details.
 */

#include "rexprivate.h"
#include "platform.h"


#define ISWORDCHAR(c) (ISALNUM(c) || (c == '_'))

#define STACK_CACHE_SIZE 1024


REX_API(void)
rexResetSearch(rex_search_t search)
{
	search_t *s;

	s = (search_t*)search;
	s->queue[0] = SEPARATOR;
	s->head = 0;
	s->tail = 0;
	s->side = 1;
	s->prevchar = '\0';
	s->base = 0;
	s->bestorigin = ~0;
	s->bestlen = 0;
	s->flags &= SEARCH_MALLOCED;

	MEMSET(s->history[0], ~0,
		sizeof(history_t) * (s->automaton->maxstate + 1) * 2);
}


static rex_search_t
PreallocSearch(rex_t rex, void *cache, size_t cachesize)
{
	size_t searchsize, qsize, histsize, qlen, size;
	int nstates;
	const automaton_t *a;
	search_t *search;

	a = (const automaton_t*)rex;
	nstates = a->maxstate + 1;
	qlen = (nstates * 2) + 1;  /* 2 queues, 1 separator */
	searchsize = ROUND_UP(sizeof(search_t), QUEUE_ALIGNMENT);
	qsize = ROUND_UP(sizeof(stateid_t) * qlen, HIST_ALIGNMENT);
	histsize = ROUND_UP(sizeof(history_t) * nstates, HIST_ALIGNMENT);
	size = searchsize + qsize + (histsize * 2);
	DPRINTF(("rexPreallocSize: %lu bytes (%s)\n",
		(unsigned long)size,
		(size > cachesize) ? "malloc'd" : "cached on stack"));
	if (size > cachesize)
	{
		search = (search_t*)MALLOC(size);
		if (search == NULL)
		{
			return REX_INVALID_SEARCH;
		}
		search->flags = SEARCH_MALLOCED;
	}
	else
	{
		search = (search_t*)cache;
		search->flags = 0;
	}
	search->queue = (stateid_t*)( (byte_t*)search + searchsize );
	search->history[0] = (history_t*)( (byte_t*)search->queue + qsize );
	search->history[1] = (history_t*)( (byte_t*)(search->history[0]) + histsize);
	search->automaton = a;
	search->size = qlen;

	/* Make sure we allocated enough space for alignment */
	ASSERT( (char*)((byte_t*)search + size) 
	        == (char*)(search->history[1] + nstates) );

	rexResetSearch((rex_search_t)search);

	return (rex_search_t)search;
}


REX_API(rex_search_t)
rexPreallocSearch(rex_t rex)
{
	return PreallocSearch(rex, NULL, 0);
}


REX_API(void)
rexFreeSearch(rex_search_t search)
{
	search_t *s;

	s = (search_t*)search;
	if (s->flags & SEARCH_MALLOCED)
	{
		FREE(s);
	}
}


REX_API(size_t)
rexMatch(rex_t rex, const REXCHAR *input, size_t len, unsigned int options,
		rex_search_t searchp, size_t *foundorigin)
{
	size_t result;
	int side;
	stateid_t *q;
	history_t **hist;
	int head;
	int tail;
	stateid_t curstate;
	size_t curorigin;
	size_t i, totali;
	size_t forigin;
	int size;
	REXCHAR c;
	size_t bestorigin;
	size_t bestlen;
	size_t starts;
	const automaton_t *a;
	search_t *search;
	byte_t stackcache[STACK_CACHE_SIZE];
	
	DPRINTF(("rexMatch input=`%s' len=%d search=%p side=%d\n",
		input, len, (void*)searchp,
		(searchp!=NULL) ? ((search_t*)searchp)->side : 69));

	/* It is mandatory to pass one of these options */
	ASSERT( (options & (REX_MATCH_WHOLE | REX_MATCH_SUBSTRING)) );
	
	if (len > REX_SEARCH_MAX)
	{
		ASSERT(FALSE);
		return 0;
	}

	a = (automaton_t*)rex;
	if (a->startstate == INVALID_BLOCKID)
	{
		if (foundorigin != NULL)
		{
			*foundorigin = 0;
		}
		/* A null regex can only match the empty string */
		return ((options & REX_MATCH_WHOLE) && (len != 0)) ? 0 : 1;
	}
	
	if (searchp == NULL)
	{
		search = (search_t*)PreallocSearch(rex,
			stackcache, sizeof(stackcache));
		options |= REX_MATCH_EOL;
	}
	else
	{
		search = (search_t*)searchp;
		if (search->flags & SEARCH_NO_MORE)
		{
			/*
			 * A REX_MATCH_WHOLE match has failed and we shouldn't
			 * even bother looking any further.
			 */
			if (foundorigin != NULL)
			{
				*foundorigin = search->bestorigin;
			}
			return search->bestlen;
		}
	}

#ifdef DEBUG
	/*
	 * You can only pass REX_MATCH_EOL on the last call
	 * to rexMatch.
	 */
	ASSERT( !(search->flags & SEARCH_NOT_ALLOWED) );
	if (options & REX_MATCH_EOL)
	{
		search->flags |= SEARCH_NOT_ALLOWED;
	}
#endif

	starts = (options & REX_MATCH_WHOLE) ?
		((search->base == 0) ? 1 : 0)
		: ~0;

	size = search->size;
	head = search->head;
	tail = search->tail;
	q = search->queue;
	hist = search->history;
	side = search->side;
	bestorigin = search->bestorigin;
	bestlen = search->bestlen;

	if (search->base > 0)
	{
		ASSERT(q[head] == SEPARATOR);
		/*
		 * therefore, side will be inverted on the first iteration of
		 * the main loop.
		 */
		MEMCPY(hist[side^1], hist[side],
				sizeof(history_t) * (a->maxstate+1));
	}

	i = ~0;
	totali = search->base - 1;
	c = search->prevchar;

	while (1)
	{
#ifdef DEBUG
		{
			int j, s;

			DPRINTF(("%34c  {", ' '));
			s = side;

			j = head - 1;
			do
			{
				j = (j + 1) % size;
				if (q[j] == SEPARATOR)
				{
					DPRINTF(("^"));
					s ^= 1;
				}
				else
				{
					DPRINTF(("%d(%d)",
						q[j], hist[s][q[j]].origin));
				}
				DPRINTF(("%c", (j==tail) ? '}' : ','));
			} while (j != tail);
			DPRINTF(("\n"));
		}
#endif

		curstate = q[head];
		if (curstate == SEPARATOR)
		{
			history_t *h;
			i++;
			totali++;

			side ^= 1;
			if (i > len)
			{
				DPRINTF(("Ran out of input%s\n",
					(head==tail) ? " and states" : ""));
				ASSERT(i == (len+1));
				/*
				 * Loop exit point 1 of 3: Run out of input
				 * At this stage a match may or may not have
				 * been found. Even if a match has been found,
				 * we can't be sure that it's the longest
				 * possible, because more input from a
				 * successive call may result in a longer match.
				 */
				break;
			}

			h = &hist[side][a->startstate];
			if ( (i < starts) && (h->offset != totali) )
			{
				h->offset = totali;
				h->origin = totali;
				q[head] = a->startstate;
				DPRINTF(("Injecting start state at origin %d\n", totali));
			}
			else
			{
				if (i < starts)
				{
					DPRINTF(("Avoided injecting start state at origin %d because a better one is queued\n", totali));
				}
				else if (head == tail)
				{
					DPRINTF(("Ran out of states\n"));
					/*
					 * Loop exit point 2 of 3: Ran out of
					 * states.
					 * No further WHOLE matching is possible, even
					 * if we are given more input.
					 */
					if (options & REX_MATCH_WHOLE)
					{
						search->flags |= SEARCH_NO_MORE;
					}
					break;
				}
				head = (head + 1) % size;
			}
			curstate = q[head];
			tail = (tail + 1) % size;
			q[tail] = SEPARATOR;
			search->prevchar = c;
			if (i < len)
			{
				c = input[i];
				if (a->options & REX_COMPILE_ICASE)
				{
					c = TOUPPER(c);
				}
			}
			else if ((searchp != NULL) && !(options & REX_MATCH_EOL))
			{
				int src;
				stateid_t x;

				DPRINTF(("All input had been consumed. Now saving side %d before exhausting the queue:", side));
				/*
				 * Make a copy of the queue on it's tail,
				 * in case the search is continued with more
				 * input.
				 */
				src = head;
				do
				{
					tail = (tail + 1) % size;
					x = q[src];
					q[tail] = x;
					src = (src + 1) % size;
					if (x != SEPARATOR)
					{
						DPRINTF((" %d", x));
					}
				} while ( x != SEPARATOR );
				tail = (tail + size - 1) % size;
				MEMCPY(hist[side^1], hist[side],
					sizeof(history_t) * (a->maxstate+1));
				DPRINTF(("\n"));
			}
		}
		ASSERT(curstate != SEPARATOR);
		head = (head + 1) % size;

		curorigin = hist[side][curstate].origin;

		DPRINTF(("state=%-2d origin=%-2d i=%-2d(%c) side=%d\n",
			curstate, curorigin, i, c, side));

		if (curstate == ACCEPT_STATE)
		{
			size_t foundlen;

			foundlen = totali - curorigin;
			DPRINTF(("Found a %ssolution: offset %d len %d\n",
				(curorigin <= bestorigin) ? "better " : "",
				curorigin, foundlen));
			/*
			 * XXX If back-references or subexpression memories
			 * are introduced, this logic needs to be improved
			 * to ensure that subexpression matches are as long
			 * as possible, left to right.
			 */
			if ( (curorigin < bestorigin) ||
			     ((curorigin == bestorigin) && (foundlen > bestlen))
			   )
			{
				if ( (options & REX_MATCH_WHOLE) ?
					(foundlen == (search->base+len)) :
					TRUE
				   )
				{
					bestorigin = curorigin;
					bestlen = foundlen;
					if (options & REX_MATCH_ANY)
					{
						/*
						 * Loop exit point 3 of 3: REX_MATCH_ANY
						 * short-circuit.
						 * The REX_MATCH_ANY option tells us that
						 * the caller does not need to know the
						 * longest match; just whether or not
						 * one exists. This allows us to drop
						 * out early in most cases.
						 */
						DPRINTF(("REX_MATCH_ANY satisfied. I'm done.\n"));
						search->flags |= SEARCH_NO_MORE;
						break;
					}
				}
			}
		}
		else
		{
			const state_t *state;
			transitionid_t tid;

			state = &a->block[curstate].state;
			tid = FIRSTTRANSID(curstate);
			while (tid != INVALID_BLOCKID)
			{
				const transition_t *trans;
				enum {NOTHING, UNSURE, APPEND, PREPEND} action;

				action = NOTHING;
				trans = &a->block[tid].trans;

				switch (trans->test)
				{
					case NOP:
						action = PREPEND;
						break;

					case START_OF_LINE:
						if ( (search->prevchar == '\n') || (totali == 0) )
						{
							action = PREPEND;
						}
						break;

					case END_OF_LINE:
						if (i < len)
						{
							if ((c == '\r') ||
							    (c == '\n'))
							{
								action = PREPEND;
							}
						}
						else if (options & REX_MATCH_EOL)
						{
							action = PREPEND;
						}
						break;

					case START_OF_WORD:
						if ( (!ISWORDCHAR(search->prevchar) || (totali == 0)) &&
						     ((i < len) && ISWORDCHAR(c))
						   )
						{
							action = PREPEND;
						}
						break;

					case END_OF_WORD:
						if ( ISWORDCHAR(search->prevchar) )
						{
							if (i < len)
							{
								if (!ISWORDCHAR(c))
								{
									action = PREPEND;
								}
							}
							else if (options & REX_MATCH_EOL)
							{
								action = PREPEND;
							}
						}
						break;

					case WORD_BOUNDARY:
						if ( (!ISWORDCHAR(search->prevchar) || (totali == 0)) &&
						     ((i < len) && ISWORDCHAR(c))
						   )
						{
							action = PREPEND;
						}
						else if ( ISWORDCHAR(search->prevchar) )
						{
							if (i < len)
							{
								if (!ISWORDCHAR(c))
								{
									action = PREPEND;
								}
							}
							else if (options & REX_MATCH_EOL)
							{
								action = PREPEND;
							}
						}
						break;

					default:
						action = UNSURE;
						break;
				}

				if ( (action == UNSURE) && (i < len) )
				{
					switch (trans->test)
					{
						case DOT:
							if (c == '\n')
							{
								action = (a->options & REX_COMPILE_NEWLINE) ?
									NOTHING : APPEND;
							}
							else
							{
								action = APPEND;
							}
							break;

						case LITERAL:
							if (c == OPERANDOF(trans)->literal)
							{
								action = APPEND;
							}
							break;

						case CHARSET:
							if ( CHARSET_CONTAINS(OPERANDOF(trans)->charset,c) )
							{
								action = APPEND;
							}
							break;

						case WHITESPACE:
							if (ISSPACE(c))
							{
								action = APPEND;
							}
							break;

						case NOT_WHITESPACE:
							if (!ISSPACE(c))
							{
								action = APPEND;
							}
							break;

						case WORD_CHAR:
							if (ISWORDCHAR(c))
							{
								action = APPEND;
							}
							break;

						case NOT_WORD_CHAR:
							if (!ISWORDCHAR(c))
							{
								action = APPEND;
							}
							break;

						case DIGIT:
							if (ISDIGIT(c))
							{
								action = APPEND;
							}
							break;

						case NOT_DIGIT:
							if (!ISDIGIT(c))
							{
								action = APPEND;
							}
							break;

						default:
							ASSERT(FALSE);
							break;
					}
				}

				if (action == APPEND)
				{
					int nexti;
					stateid_t dest;
					history_t *p;

					dest = trans->dest;
					nexti = totali + 1;

					p = &hist[side^1][dest];
					if (p->offset != nexti)
					{
#ifdef DEBUG
						/* assert that dest is not already in the back-queue */
						{
							int j;

							j = tail;
							while (q[j] != SEPARATOR)
							{
								ASSERT( q[j] != dest );
								j = (j + size - 1) % size;
							}
						}
#endif
						tail = (tail + 1) % size;
						ASSERT(head != tail); /* overflow */
						p->offset = nexti;
						p->origin = curorigin;
						q[tail] = dest;
						/*
						DPRINTF(("hist[%d][%d] = %d\n", side^1, dest, nexti));
						*/
					}
					else
					{
						DPRINTF(("avoided appending state %d (hist[%d][%d]==%d)", dest, side^1, dest, nexti));
						if (curorigin < p->origin)
						{
							DPRINTF((", but took lower origin: was %d now %d", p->origin, curorigin));
							p->origin = curorigin;
						}
						DPRINTF(("\n"));
					}
				}
				else if (action == PREPEND)
				{
					stateid_t dest;
					history_t *p;

					dest = trans->dest;

					p = &hist[side][dest];
					if ( (p->offset != totali) || (curorigin < p->origin) )
					{
#ifdef DEBUG
						/* assert that dest is not already in the front-queue */
						{
							int j;

							j = head;
							while (q[j] != SEPARATOR)
							{
								ASSERT( q[j] != dest );
								j = (j + 1) % size;
							}
						}
#endif
						/*
						 * Proof that this code won't overflow the front-queue:
						 * 1. Start state insertions are always done before
						 *    transition traversal (see above).
						 * 2. Therefore, states on the front-queue are in
						 *    descending order of origin.
						 * 3. Therefore, the state "dest" front-queued with
						 *    origin "p->origin" must have been visited already
						 *    (nolonger on the front-queue).
						 */

						head = (head + size - 1) % size;
						ASSERT(head != tail); /* overflow */
						p->offset = totali;
						p->origin = curorigin;
						q[head] = dest;
						/*
						DPRINTF(("hist[%d][%d] = %d\n", side, dest, totali));
						*/
					}
					else
					{
						DPRINTF(("avoided prepending state %d (hist[%d][%d]==%d)\n", dest, side, dest, totali));
					}
				}

				tid = trans->sibling;
			}
		}
	}

	DPRINTF(("Match: bestorigin=%d bestlen=%d\n",
		bestorigin, bestlen));

	result = (bestorigin != ~0) ?
		(options & REX_MATCH_WHOLE) ?
			(bestlen == search->base+len) ? (bestlen+1) : 0 :
			(bestlen+1)
		:
			0;

	forigin = result ? bestorigin : totali-1;
	if (foundorigin != NULL)
	{
		*foundorigin = forigin;
	}
			
	if (searchp == NULL)
	{
		rexFreeSearch(search);
	}
	else
	{
		search->head = head;
		search->tail = tail;
		search->side = side;
		search->base += len;
		if (search->flags & SEARCH_NO_MORE)
		{
			search->bestlen = result;
			search->bestorigin = forigin;
		}
		else
		{
			search->bestlen = bestlen;
			search->bestorigin = bestorigin;
		}
	}

	return result;
}

